from unittest.mock import AsyncMock, MagicMock
from astromesh.mcp.client import MCPClient, MCPToolInfo
from astromesh.mcp.server import MCPServer
from astromesh.core.tools import ToolRegistry


async def test_mcp_client_call_tool():
    client = MCPClient({"transport": "http", "url": "http://localhost:3000/mcp"})
    client._http_client = AsyncMock()
    mock_response = MagicMock()
    mock_response.json.return_value = {
        "result": {"content": [{"type": "text", "text": "tool result"}]}
    }
    mock_response.raise_for_status = MagicMock()
    client._http_client.post = AsyncMock(return_value=mock_response)

    result = await client.call_tool("test_tool", {"arg": "value"})
    assert result["result"] == "tool result"


async def test_mcp_client_get_tools():
    client = MCPClient({"transport": "http", "url": "http://localhost:3000/mcp"})
    client._tools = [MCPToolInfo(name="tool1", description="A tool", parameters={})]
    tools = client.get_tools()
    assert len(tools) == 1
    assert tools[0].name == "tool1"


async def test_mcp_server_tools_list():
    from httpx import AsyncClient, ASGITransport
    from fastapi import FastAPI

    runtime = MagicMock()
    runtime.list_agents.return_value = [
        {"name": "echo", "version": "0.1.0", "namespace": "default"}
    ]

    app = FastAPI()
    mcp_server = MCPServer(runtime=runtime)
    app.include_router(mcp_server.router)

    transport = ASGITransport(app=app)
    async with AsyncClient(transport=transport, base_url="http://test") as client:
        resp = await client.post("/mcp", json={
            "jsonrpc": "2.0", "id": 1, "method": "tools/list", "params": {}
        })
        assert resp.status_code == 200
        data = resp.json()
        assert len(data["result"]["tools"]) == 1
        assert data["result"]["tools"][0]["name"] == "agent_echo"


async def test_tool_registry_register_mcp():
    registry = ToolRegistry()
    mock_client = MagicMock()
    mock_client.get_tools.return_value = [
        MCPToolInfo(name="mcp_tool", description="An MCP tool", parameters={"x": {"type": "number"}})
    ]
    count = await registry.register_mcp_server("test-server", mock_client)
    assert count == 1
    assert "mcp_tool" in registry._tools
    schemas = registry.get_tool_schemas()
    assert any(s["function"]["name"] == "mcp_tool" for s in schemas)



"""Wizard presets for Orbit deployments."""

from __future__ import annotations

from typing import Any

PRESETS: dict[str, dict[str, Any]] = {
    "starter": {
        "estimated_cost": 15,
        "compute": {
            "runtime": {"min_instances": 1, "max_instances": 3, "cpu": "2", "memory": "2Gi"},
        },
        "database": {
            "tier": "db-f1-micro",
            "version": "POSTGRES_16",
            "storage_gb": 10,
            "high_availability": False,
        },
        "cache": {"tier": "basic", "memory_gb": 1},
        "secrets": {"provider_keys": True, "jwt_secret": True},
        "images": {
            "runtime": "fulfarodev/astromesh:latest",
        },
    },
    "pro": {
        "estimated_cost": 80,
        "compute": {
            "runtime": {"min_instances": 1, "max_instances": 5, "cpu": "4", "memory": "4Gi"},
        },
        "database": {
            "tier": "db-g1-small",
            "version": "POSTGRES_16",
            "storage_gb": 20,
            "high_availability": True,
        },
        "cache": {"tier": "standard", "memory_gb": 4},
        "secrets": {"provider_keys": True, "jwt_secret": True},
        "images": {
            "runtime": "fulfarodev/astromesh:latest",
        },
    },
}


def build_orbit_yaml(
    *,
    name: str,
    environment: str,
    provider: str,
    project: str,
    region: str,
    preset: str,
) -> dict:
    """Build a complete orbit.yaml dict from wizard answers and a preset."""
    p = PRESETS[preset]
    return {
        "apiVersion": "astromesh/v1",
        "kind": "OrbitDeployment",
        "metadata": {"name": name, "environment": environment},
        "spec": {
            "provider": {"name": provider, "project": project, "region": region},
            "compute": p["compute"],
            "database": p["database"],
            "cache": p["cache"],
            "secrets": p["secrets"],
            "images": p["images"],
        },
    }



import json
from datetime import datetime

import aiosqlite

from astromesh.core.memory import ConversationBackend, ConversationTurn


class SQLiteConversationBackend(ConversationBackend):
    def __init__(self, db_path: str):
        self._db_path = db_path
        self._db: aiosqlite.Connection | None = None

    async def initialize(self):
        self._db = await aiosqlite.connect(self._db_path)
        await self._db.execute(
            "CREATE TABLE IF NOT EXISTS conversations ("
            "id INTEGER PRIMARY KEY AUTOINCREMENT, "
            "session_id TEXT, "
            "role TEXT, "
            "content TEXT, "
            "metadata TEXT, "
            "token_count INTEGER, "
            "timestamp TEXT"
            ")"
        )
        await self._db.execute(
            "CREATE TABLE IF NOT EXISTS summaries (session_id TEXT PRIMARY KEY, summary TEXT)"
        )
        await self._db.commit()

    async def save_turn(self, session_id, turn):
        await self._db.execute(
            "INSERT INTO conversations (session_id, role, content, metadata, token_count, timestamp) "
            "VALUES (?, ?, ?, ?, ?, ?)",
            (
                session_id,
                turn.role,
                turn.content,
                json.dumps(turn.metadata),
                turn.token_count,
                turn.timestamp.isoformat(),
            ),
        )
        await self._db.commit()

    async def get_history(self, session_id, limit=50):
        cursor = await self._db.execute(
            "SELECT role, content, metadata, token_count, timestamp "
            "FROM conversations WHERE session_id = ? ORDER BY id DESC LIMIT ?",
            (session_id, limit),
        )
        rows = await cursor.fetchall()
        rows.reverse()
        return [
            ConversationTurn(
                role=row[0],
                content=row[1],
                timestamp=datetime.fromisoformat(row[4]),
                metadata=json.loads(row[2]) if row[2] else {},
                token_count=row[3] or 0,
            )
            for row in rows
        ]

    async def clear(self, session_id):
        await self._db.execute("DELETE FROM conversations WHERE session_id = ?", (session_id,))
        await self._db.execute("DELETE FROM summaries WHERE session_id = ?", (session_id,))
        await self._db.commit()

    async def get_summary(self, session_id):
        cursor = await self._db.execute(
            "SELECT summary FROM summaries WHERE session_id = ?", (session_id,)
        )
        row = await cursor.fetchone()
        return row[0] if row else None

    async def save_summary(self, session_id, summary):
        await self._db.execute(
            "INSERT OR REPLACE INTO summaries (session_id, summary) VALUES (?, ?)",
            (session_id, summary),
        )
        await self._db.commit()



import pytest
from astromesh.core.tools import ToolRegistry, ToolType


@pytest.fixture
def registry():
    return ToolRegistry()


async def double_handler(x: int):
    return {"result": x * 2}


def test_register_internal_tool(registry):
    registry.register_internal(
        name="double",
        handler=double_handler,
        description="Doubles a number",
        parameters={"type": "object", "properties": {"x": {"type": "integer"}}},
    )
    assert "double" in registry._tools
    assert registry._tools["double"].tool_type == ToolType.INTERNAL
    assert registry._tools["double"].handler is double_handler


@pytest.mark.asyncio
async def test_execute_internal_tool(registry):
    registry.register_internal(
        name="double",
        handler=double_handler,
        description="Doubles a number",
        parameters={"type": "object", "properties": {"x": {"type": "integer"}}},
    )
    result = await registry.execute("double", {"x": 5})
    assert result == {"result": 10}


@pytest.mark.asyncio
async def test_execute_missing_tool(registry):
    result = await registry.execute("nonexistent", {})
    assert "error" in result
    assert "not found" in result["error"]


def test_get_tool_schemas(registry):
    registry.register_internal(
        name="tool_a",
        handler=double_handler,
        description="Tool A",
        parameters={"type": "object", "properties": {"x": {"type": "integer"}}},
    )
    registry.register_internal(
        name="tool_b",
        handler=double_handler,
        description="Tool B",
        parameters={"type": "object", "properties": {"y": {"type": "string"}}},
    )
    schemas = registry.get_tool_schemas()
    assert len(schemas) == 2
    assert schemas[0]["type"] == "function"
    assert "function" in schemas[0]
    assert schemas[0]["function"]["name"] == "tool_a"
    assert schemas[1]["function"]["name"] == "tool_b"
    assert "parameters" in schemas[0]["function"]


def test_get_tool_schemas_filtered(registry):
    registry.register_internal(
        name="admin_tool",
        handler=double_handler,
        description="Admin only",
        parameters={},
        permissions=["admin"],
    )
    registry.register_internal(
        name="public_tool",
        handler=double_handler,
        description="Public tool",
        parameters={},
    )
    # Agent with "user" permission should not see admin_tool but should see public_tool
    schemas = registry.get_tool_schemas(agent_permissions=["user"])
    names = [s["function"]["name"] for s in schemas]
    assert "public_tool" in names
    assert "admin_tool" not in names

    # Agent with "admin" permission should see both
    schemas_admin = registry.get_tool_schemas(agent_permissions=["admin"])
    names_admin = [s["function"]["name"] for s in schemas_admin]
    assert "admin_tool" in names_admin
    assert "public_tool" in names_admin



"""System management endpoints for Astromesh Node."""

import os
import time

from fastapi import APIRouter
from pydantic import BaseModel

from astromesh import __version__

router = APIRouter(prefix="/system", tags=["system"])

_runtime = None
_start_time = time.time()


def set_runtime(runtime):
    global _runtime
    _runtime = runtime


class CheckResult(BaseModel):
    status: str
    message: str = ""


class StatusResponse(BaseModel):
    version: str
    uptime_seconds: float
    mode: str
    agents_loaded: int
    pid: int
    services: dict[str, bool] = {}
    peers: list[dict] = []
    mesh: dict | None = None


class DoctorResponse(BaseModel):
    healthy: bool
    checks: dict[str, CheckResult]


def _detect_mode() -> str:
    if os.path.exists("/etc/astromesh/runtime.yaml"):
        return "system"
    return "dev"


@router.get("/status", response_model=StatusResponse)
async def system_status():
    agents_loaded = 0
    services = {}
    peers = []
    if _runtime:
        agents_loaded = len(_runtime.list_agents())
        if hasattr(_runtime, "service_manager") and _runtime.service_manager:
            services = _runtime.service_manager.to_dict()
        if hasattr(_runtime, "peer_client") and _runtime.peer_client:
            peers = _runtime.peer_client.to_dict()

    mesh_info = None
    if _runtime and hasattr(_runtime, "mesh_manager") and _runtime.mesh_manager:
        mm = _runtime.mesh_manager
        mesh_info = {
            "enabled": True,
            "node_id": mm.node_id,
            "node_name": mm._config.node_name,
            "leader": mm.cluster_state().leader_id,
            "cluster_size": len(mm.cluster_state().nodes),
            "status": mm.local_node_state().status,
        }

    return StatusResponse(
        version=__version__,
        uptime_seconds=round(time.time() - _start_time, 2),
        mode=_detect_mode(),
        agents_loaded=agents_loaded,
        pid=os.getpid(),
        services=services,
        peers=peers,
        mesh=mesh_info,
    )


@router.get("/doctor", response_model=DoctorResponse)
async def system_doctor():
    checks: dict[str, CheckResult] = {}

    if not _runtime:
        checks["runtime"] = CheckResult(status="unavailable", message="Runtime not initialized")
        return DoctorResponse(healthy=False, checks=checks)

    checks["runtime"] = CheckResult(status="ok", message="Runtime initialized")

    providers_checked = set()
    for agent in _runtime._agents.values():
        if hasattr(agent, "_router") and hasattr(agent._router, "_providers"):
            for name, provider in agent._router._providers.items():
                if name not in providers_checked:
                    providers_checked.add(name)
                    try:
                        healthy = await provider.health_check()
                        checks[f"provider:{name}"] = CheckResult(
                            status="ok" if healthy else "degraded",
                            message=f"Provider {name} health check",
                        )
                    except Exception as e:
                        checks[f"provider:{name}"] = CheckResult(status="error", message=str(e))

    # Check peers
    if hasattr(_runtime, "peer_client") and _runtime.peer_client:
        for peer in _runtime.peer_client.list_peers():
            healthy = await _runtime.peer_client.health_check(peer["name"])
            checks[f"peer:{peer['name']}"] = CheckResult(
                status="ok" if healthy else "unreachable",
                message=f"Peer {peer['name']} at {peer['url']}",
            )

    all_ok = all(c.status == "ok" for c in checks.values())
    return DoctorResponse(healthy=all_ok, checks=checks)



def test_all_public_exports():
    """All documented public API names are importable from the top-level package."""
    from astromesh_adk import (
        Agent,
        AgentTeam,
        ADKRuntime,
        Callbacks,
        RunContext,
        RunResult,
        StreamEvent,
        Tool,
        ToolContext,
        agent,
        connect,
        disconnect,
        mcp_tools,
        remote,
        tool,
    )

    assert Agent is not None
    assert agent is not None
    assert tool is not None
    assert Tool is not None
    assert connect is not None
    assert disconnect is not None
    assert remote is not None
    assert AgentTeam is not None
    assert ADKRuntime is not None
    assert Callbacks is not None
    assert RunResult is not None
    assert RunContext is not None
    assert StreamEvent is not None
    assert ToolContext is not None
    assert mcp_tools is not None


def test_version():
    import astromesh_adk
    assert hasattr(astromesh_adk, "__version__")
    assert astromesh_adk.__version__ == "0.1.0"



"""Tests for the CLI main entry point."""

from astromesh_cli.main import app


def test_app_name():
    assert app.info.name == "astromeshctl"


def test_app_has_commands():
    """Check that key commands are registered."""
    command_names = [cmd.name for cmd in app.registered_commands]
    callback_names = [
        cmd.callback.__name__ for cmd in app.registered_commands if cmd.callback
    ]
    group_names = [g.name for g in app.registered_groups]

    # Typer groups (sub-apps)
    assert "status" in group_names
    assert "agents" in group_names
    assert "providers" in group_names
    assert "traces" in group_names
    assert "tools" in group_names

    # Direct commands
    assert "run" in command_names
    assert "dev" in command_names
    assert "ask" in command_names
    assert "metrics" in command_names

    # version is registered via @app.command() decorator — name defaults to None
    # but its callback function is named "version"
    assert "version" in callback_names


def test_app_has_no_args_is_help():
    """App should show help when called with no arguments."""
    assert app.info.no_args_is_help is True



"""astromeshctl traces/trace — Observability trace commands."""

import typer

from astromesh_cli.client import api_get, api_get_params
from astromesh_cli.output import (
    console,
    print_error,
    print_json,
    print_trace_list,
    print_trace_tree,
)

app = typer.Typer(help="View execution traces.")


@app.command("list")
def traces_list_command(
    agent: str = typer.Argument(..., help="Agent name to filter traces"),
    last: int = typer.Option(10, "--last", help="Number of recent traces to show"),
    json_output: bool = typer.Option(False, "--json", help="Output raw JSON"),
) -> None:
    """List recent traces for an agent."""
    try:
        data = api_get_params("/v1/traces/", params={"agent": agent, "limit": last})
    except Exception as e:
        print_error(f"Failed to fetch traces: {e}")
        raise typer.Exit(code=1)

    if json_output:
        print_json(data)
        return

    traces = data.get("traces", [])
    if not traces:
        console.print("[dim]No traces found.[/dim]")
        return

    print_trace_list(traces)


def trace_command(
    trace_id: str = typer.Argument(..., help="Trace ID to inspect"),
    json_output: bool = typer.Option(False, "--json", help="Output raw JSON"),
) -> None:
    """Show detailed trace with span tree."""
    try:
        data = api_get(f"/v1/traces/{trace_id}")
    except Exception as e:
        print_error(f"Failed to fetch trace: {e}")
        raise typer.Exit(code=1)

    if json_output:
        print_json(data)
        return

    print_trace_tree(data)



"""Tests for /v1/system/* API endpoints."""

from unittest.mock import AsyncMock, MagicMock

import pytest

async def test_system_status_no_runtime(client):
    resp = await client.get("/v1/system/status")
    assert resp.status_code == 200
    data = resp.json()
    assert "version" in data
    assert "uptime_seconds" in data
    assert data["mode"] in ("dev", "system")
    assert "agents_loaded" in data


async def test_system_status_with_runtime(client):
    from astromesh.api.routes import system

    mock_runtime = MagicMock()
    mock_runtime.list_agents.return_value = [{"name": "agent1"}, {"name": "agent2"}]
    mock_runtime.service_manager = None
    mock_runtime.peer_client = None
    system.set_runtime(mock_runtime)

    resp = await client.get("/v1/system/status")
    assert resp.status_code == 200
    data = resp.json()
    assert data["agents_loaded"] == 2
    system.set_runtime(None)


async def test_system_doctor_no_runtime(client):
    from astromesh.api.routes import system

    system.set_runtime(None)

    resp = await client.get("/v1/system/doctor")
    assert resp.status_code == 200
    data = resp.json()
    assert data["healthy"] is False
    assert data["checks"]["runtime"]["status"] == "unavailable"


async def test_system_doctor_with_runtime(client):
    from astromesh.api.routes import system

    mock_runtime = MagicMock()
    mock_runtime.list_agents.return_value = [{"name": "a1"}]
    mock_runtime._agents = {"a1": MagicMock()}

    mock_provider = AsyncMock()
    mock_provider.health_check = AsyncMock(return_value=True)
    mock_runtime._agents["a1"]._router._providers = {"ollama": mock_provider}

    system.set_runtime(mock_runtime)

    resp = await client.get("/v1/system/doctor")
    assert resp.status_code == 200
    data = resp.json()
    assert data["checks"]["runtime"]["status"] == "ok"
    system.set_runtime(None)


async def test_system_status_includes_services(client):
    from astromesh.api.routes import system
    from astromesh.runtime.services import ServiceManager

    mock_runtime = MagicMock()
    mock_runtime.list_agents.return_value = []
    mock_runtime.service_manager = ServiceManager({"inference": False})
    mock_runtime.peer_client = None
    system.set_runtime(mock_runtime)

    resp = await client.get("/v1/system/status")
    assert resp.status_code == 200
    data = resp.json()
    assert "services" in data
    assert data["services"]["inference"] is False
    assert data["services"]["api"] is True
    system.set_runtime(None)


async def test_system_status_includes_peers(client):
    from astromesh.api.routes import system
    from astromesh.runtime.peers import PeerClient

    mock_runtime = MagicMock()
    mock_runtime.list_agents.return_value = []
    mock_runtime.service_manager = None
    mock_runtime.peer_client = PeerClient(
        [
            {"name": "inference-1", "url": "http://inference:8000", "services": ["inference"]},
        ]
    )
    system.set_runtime(mock_runtime)

    resp = await client.get("/v1/system/status")
    assert resp.status_code == 200
    data = resp.json()
    assert "peers" in data
    assert len(data["peers"]) == 1
    assert data["peers"][0]["name"] == "inference-1"
    system.set_runtime(None)


async def test_system_doctor_checks_peers(client):
    from astromesh.api.routes import system
    from astromesh.runtime.peers import PeerClient

    mock_runtime = MagicMock()
    mock_runtime.list_agents.return_value = []
    mock_runtime._agents = {}
    mock_runtime.service_manager = None

    pc = PeerClient(
        [
            {"name": "inference-1", "url": "http://inference:8000", "services": ["inference"]},
        ]
    )
    pc.health_check = AsyncMock(return_value=False)
    mock_runtime.peer_client = pc
    system.set_runtime(mock_runtime)

    resp = await client.get("/v1/system/doctor")
    assert resp.status_code == 200
    data = resp.json()
    assert "peer:inference-1" in data["checks"]
    system.set_runtime(None)



import pytest
from astromesh_adk.tools import tool, Tool, ToolDefinitionWrapper


# --- @tool decorator tests ---

@tool(description="Search the web")
async def web_search(query: str, max_results: int = 5) -> str:
    return f"Results for {query}"


@tool(description="Calculate", rate_limit={"max_calls": 10, "window_seconds": 60})
async def calculator(expression: str) -> float:
    return 42.0


def test_tool_decorator_preserves_callable():
    """The decorated function is still callable as a tool handler."""
    assert callable(web_search)


def test_tool_decorator_creates_wrapper():
    assert isinstance(web_search, ToolDefinitionWrapper)


def test_tool_decorator_name_from_function():
    assert web_search.tool_name == "web_search"


def test_tool_decorator_description():
    assert web_search.tool_description == "Search the web"


def test_tool_decorator_schema_generation():
    schema = web_search.parameters_schema
    assert schema["type"] == "object"
    assert "query" in schema["properties"]
    assert schema["properties"]["query"]["type"] == "string"
    assert "max_results" in schema["properties"]
    assert schema["properties"]["max_results"]["type"] == "integer"
    assert schema["properties"]["max_results"].get("default") == 5
    assert schema["required"] == ["query"]


def test_tool_decorator_rate_limit():
    assert calculator.rate_limit == {"max_calls": 10, "window_seconds": 60}


async def test_tool_decorator_execution():
    result = await web_search(query="test", max_results=3)
    assert result == "Results for test"


# --- Optional type hints ---

@tool(description="Optional param")
async def optional_tool(name: str, count: int | None = None) -> str:
    return name


def test_tool_optional_params():
    schema = optional_tool.parameters_schema
    assert "count" in schema["properties"]
    assert "count" not in schema["required"]


# --- Tool base class tests ---

class MyTool(Tool):
    name = "my_tool"
    description = "A custom tool"

    def parameters(self):
        return {
            "input": {"type": "string", "description": "The input"},
        }

    async def execute(self, args: dict, ctx=None) -> str:
        return f"processed: {args['input']}"


async def test_tool_class_execute():
    t = MyTool()
    result = await t.execute({"input": "hello"})
    assert result == "processed: hello"


def test_tool_class_attributes():
    t = MyTool()
    assert t.name == "my_tool"
    assert t.description == "A custom tool"
    assert "input" in t.parameters()



"""Guardrails configuration normalization."""

from __future__ import annotations

# Default actions per guardrail type
DEFAULT_ACTIONS = {
    "pii_detection": "redact",
    "topic_filter": "block",
    "max_length": "block",
    "cost_limit": "block",
    "content_filter": "redact",
}


def normalize_guardrails_config(guardrails: dict | None) -> dict:
    """Normalize guardrails config into Astromesh format.

    Accepts:
        - None → {"input": [], "output": []}
        - {"input": ["pii_detection"]} → {"input": [{"type": "pii_detection", "action": "redact"}]}
        - {"input": [{"type": "pii_detection", "action": "warn"}]} → passthrough
    """
    if guardrails is None:
        return {"input": [], "output": []}

    result = {}
    for side in ("input", "output"):
        items = guardrails.get(side, [])
        normalized = []
        for item in items:
            if isinstance(item, str):
                normalized.append({
                    "type": item,
                    "action": DEFAULT_ACTIONS.get(item, "block"),
                })
            elif isinstance(item, dict):
                normalized.append(item)
        result[side] = normalized

    return result



"""Tests for daemon core — argument parsing."""

from astromesh_node.daemon.core import parse_args


def test_parse_args_defaults():
    args = parse_args([])
    assert args.config is None
    assert args.host is None
    assert args.port is None
    assert args.log_level == "info"
    assert args.foreground is False


def test_parse_args_foreground():
    args = parse_args(["--foreground"])
    assert args.foreground is True


def test_parse_args_config():
    args = parse_args(["--config", "/etc/astromesh"])
    assert args.config == "/etc/astromesh"



import os
from pathlib import Path

import yaml
from fastapi import APIRouter, HTTPException

router = APIRouter()
_templates_dir: str | None = None


def set_templates_dir(path: str | None) -> None:
    """Override template directory (tests). None = use auto-discovery."""
    global _templates_dir
    _templates_dir = path


def _bundled_templates_root() -> Path | None:
    """Shipped / repo `config/templates` next to the installed package."""
    try:
        import astromesh

        p = (Path(astromesh.__file__).resolve().parent.parent / "config" / "templates")
        return p if p.is_dir() else None
    except Exception:
        return None


def _iter_template_root_dirs() -> list[Path]:
    """Directories to scan for `*.template.yaml`, in merge order (later overrides name)."""
    if _templates_dir is not None:
        p = Path(_templates_dir)
        return [p.resolve()] if p.is_dir() else []

    seen: set[Path] = set()
    roots: list[Path] = []

    def add(path: Path | None) -> None:
        if path is None or not path.is_dir():
            return
        r = path.resolve()
        if r in seen:
            return
        seen.add(r)
        roots.append(r)

    add(_bundled_templates_root())
    cfg = os.environ.get("ASTROMESH_CONFIG_DIR", "").strip()
    if cfg:
        add(Path(cfg).expanduser().resolve() / "templates")
    add((Path.cwd() / "config" / "templates").resolve())
    env = os.environ.get("ASTROMESH_TEMPLATES_DIR", "").strip()
    if env:
        add(Path(env).expanduser().resolve())
    return roots


def _load_templates() -> list[dict]:
    by_name: dict[str, dict] = {}
    for root in _iter_template_root_dirs():
        for f in sorted(root.glob("*.template.yaml")):
            with open(f, encoding="utf-8") as fh:
                data = yaml.safe_load(fh)
            if not isinstance(data, dict):
                continue
            meta = data.get("metadata") or {}
            tname = meta.get("name")
            if not tname:
                continue
            by_name[str(tname)] = data
    return list(by_name.values())


@router.get("/templates")
async def list_templates():
    templates = _load_templates()
    return [
        {
            "name": t["metadata"]["name"],
            "version": t["metadata"].get("version", ""),
            "category": t["metadata"].get("category", ""),
            "tags": t["metadata"].get("tags", []),
            "display_name": t["template"]["display_name"],
            "description": t["template"]["description"],
            "recommended_channels": t["template"].get("recommended_channels", []),
        }
        for t in templates
    ]


@router.get("/templates/{template_name}")
async def get_template(template_name: str):
    templates = _load_templates()
    for t in templates:
        if t["metadata"]["name"] == template_name:
            return {
                "name": t["metadata"]["name"],
                "version": t["metadata"].get("version", ""),
                "category": t["metadata"].get("category", ""),
                "tags": t["metadata"].get("tags", []),
                "display_name": t["template"]["display_name"],
                "description": t["template"]["description"],
                "recommended_channels": t["template"].get("recommended_channels", []),
                "variables": t["template"].get("variables", []),
                "agent_config": t["template"]["agent_config"],
            }
    raise HTTPException(404, f"Template '{template_name}' not found")



from fastapi import APIRouter

router = APIRouter(prefix="/metrics", tags=["observability"])

# In-memory counters (simple approach, later backed by prometheus-client)
_counters: dict[str, int] = {}
_histograms: dict[str, list[float]] = {}


def increment(name: str, value: int = 1):
    _counters[name] = _counters.get(name, 0) + value


def observe(name: str, value: float):
    _histograms.setdefault(name, []).append(value)


def get_counters() -> dict[str, int]:
    return dict(_counters)


def get_histograms() -> dict[str, dict]:
    result = {}
    for name, values in _histograms.items():
        if values:
            result[name] = {
                "count": len(values),
                "sum": sum(values),
                "avg": sum(values) / len(values),
                "min": min(values),
                "max": max(values),
            }
    return result


@router.get("/")
async def get_metrics():
    return {"counters": get_counters(), "histograms": get_histograms()}


@router.post("/reset")
async def reset_metrics():
    _counters.clear()
    _histograms.clear()
    return {"status": "reset"}



"""Typed resource specifications for Orbit deployments."""

from dataclasses import dataclass


@dataclass(frozen=True)
class ComputeSpec:
    min_instances: int = 1
    max_instances: int = 3
    cpu: str = "1"
    memory: str = "1Gi"


@dataclass(frozen=True)
class DatabaseSpec:
    tier: str = "db-f1-micro"
    version: str = "POSTGRES_16"
    storage_gb: int = 10
    high_availability: bool = False


@dataclass(frozen=True)
class CacheSpec:
    tier: str = "basic"
    memory_gb: int = 1


@dataclass(frozen=True)
class SecretsSpec:
    provider_keys: bool = True
    jwt_secret: bool = True


@dataclass(frozen=True)
class ImagesSpec:
    runtime: str = "fulfarodev/astromesh:latest"



"""Tests for WindowsInstaller."""

from unittest.mock import patch

from astromesh_node.installer.base import InstallerProtocol
from astromesh_node.installer.windows import WindowsInstaller


def test_implements_protocol():
    assert isinstance(WindowsInstaller(), InstallerProtocol)


def test_config_dir():
    with patch("astromesh_node.installer.windows.os") as mock_os:
        mock_os.environ = {"ProgramData": "C:\\ProgramData"}
        installer = WindowsInstaller()
        result = installer.config_dir()
        # On Linux, backslashes in "C:\ProgramData" are NOT path separators,
        # so Path treats it as a single component. Compare last 2 parts only.
        assert result.parts[-2:] == ("Astromesh", "config")
        assert "ProgramData" in str(result)


def test_bin_dir():
    with patch("astromesh_node.installer.windows.os") as mock_os:
        mock_os.environ = {"ProgramFiles": "C:\\Program Files"}
        installer = WindowsInstaller()
        result = installer.bin_dir()
        assert result.parts[-2:] == ("Astromesh", "venv")
        assert "Program Files" in str(result)



import uuid

import numpy as np
import faiss

from astromesh.core.memory import SemanticBackend, SemanticMemory


class FAISSSemanticBackend(SemanticBackend):
    def __init__(self, dimension: int = 1536):
        self._dimension = dimension
        self._indices: dict[str, faiss.IndexFlatIP] = {}
        self._data: dict[str, list[dict]] = {}

    def _ensure_index(self, agent_id: str):
        if agent_id not in self._indices:
            self._indices[agent_id] = faiss.IndexFlatIP(self._dimension)
            self._data[agent_id] = []

    async def store(self, agent_id, content, embedding, metadata):
        self._ensure_index(agent_id)
        memory_id = str(uuid.uuid4())
        vec = np.array([embedding], dtype=np.float32)
        # Normalize for cosine similarity with IndexFlatIP
        faiss.normalize_L2(vec)
        self._indices[agent_id].add(vec)
        self._data[agent_id].append(
            {
                "id": memory_id,
                "content": content,
                "embedding": embedding,
                "metadata": metadata,
            }
        )
        return memory_id

    async def search(self, agent_id, query_embedding, top_k=10, threshold=0.7):
        self._ensure_index(agent_id)
        if self._indices[agent_id].ntotal == 0:
            return []

        query_vec = np.array([query_embedding], dtype=np.float32)
        faiss.normalize_L2(query_vec)
        k = min(top_k, self._indices[agent_id].ntotal)
        scores, indices = self._indices[agent_id].search(query_vec, k)

        results = []
        for score, idx in zip(scores[0], indices[0]):
            if idx < 0 or float(score) < threshold:
                continue
            entry = self._data[agent_id][idx]
            results.append(
                SemanticMemory(
                    content=entry["content"],
                    embedding=entry["embedding"],
                    metadata=entry["metadata"],
                    similarity=float(score),
                    source=entry["id"],
                )
            )
        return results

    async def delete(self, agent_id, memory_id):
        if agent_id not in self._data:
            return
        # Find and remove the entry, then rebuild the index
        new_data = [e for e in self._data[agent_id] if e["id"] != memory_id]
        if len(new_data) == len(self._data[agent_id]):
            return  # Not found

        self._data[agent_id] = new_data
        self._indices[agent_id] = faiss.IndexFlatIP(self._dimension)
        if new_data:
            vecs = np.array([e["embedding"] for e in new_data], dtype=np.float32)
            faiss.normalize_L2(vecs)
            self._indices[agent_id].add(vecs)



import pytest

from astromesh.tools import ToolLoader
from astromesh.tools.base import BuiltinTool, ToolContext, ToolResult


class TestToolResult:
    def test_success_result(self):
        result = ToolResult(success=True, data={"key": "value"}, metadata={})
        assert result.success is True
        assert result.data == {"key": "value"}
        assert result.error is None

    def test_error_result(self):
        result = ToolResult(success=False, data=None, metadata={}, error="Something failed")
        assert result.success is False
        assert result.error == "Something failed"

    def test_to_dict(self):
        result = ToolResult(success=True, data="hello", metadata={"tokens": 10})
        d = result.to_dict()
        assert d == {"success": True, "data": "hello", "metadata": {"tokens": 10}, "error": None}

    def test_to_dict_backward_compat(self):
        result = ToolResult(success=True, data={"answer": "42"}, metadata={})
        d = result.to_dict()
        assert isinstance(d, dict)
        assert d["data"]["answer"] == "42"


class TestToolContext:
    def test_context_has_required_fields(self):
        ctx = ToolContext(
            agent_name="test-agent",
            session_id="sess-123",
            trace_span=None,
            cache={},
            secrets={},
        )
        assert ctx.agent_name == "test-agent"
        assert ctx.session_id == "sess-123"
        assert ctx.trace_span is None


class _FakeTool(BuiltinTool):
    name = "fake_tool"
    description = "A fake tool for testing"
    parameters = {"type": "object", "properties": {"query": {"type": "string"}}}

    async def execute(self, arguments: dict, context: ToolContext) -> ToolResult:
        return ToolResult(success=True, data=f"fake: {arguments.get('query')}", metadata={})


class TestToolLoader:
    def test_register_and_get(self):
        loader = ToolLoader()
        loader.register_class(_FakeTool)
        assert loader.get("fake_tool") is _FakeTool

    def test_get_unknown_returns_none(self):
        loader = ToolLoader()
        assert loader.get("nonexistent") is None

    def test_list_available(self):
        loader = ToolLoader()
        loader.register_class(_FakeTool)
        available = loader.list_available()
        assert "fake_tool" in available

    async def test_create_instance_with_config(self):
        loader = ToolLoader()
        loader.register_class(_FakeTool)
        instance = loader.create("fake_tool", config={"key": "val"})
        assert isinstance(instance, _FakeTool)
        assert instance.config == {"key": "val"}

    def test_create_unknown_raises(self):
        loader = ToolLoader()
        with pytest.raises(KeyError, match="nonexistent"):
            loader.create("nonexistent")



"""Leader election using bully algorithm."""

from __future__ import annotations

import logging

from astromesh.mesh.manager import MeshManager

logger = logging.getLogger("astromesh.mesh.leader")


class LeaderElector:
    """Bully algorithm leader election — highest node ID wins."""

    def __init__(self, mesh: MeshManager):
        self._mesh = mesh
        self._leader_id: str | None = None

    def current_leader(self) -> str | None:
        return self._leader_id

    def is_leader(self) -> bool:
        return self._leader_id == self._mesh.node_id

    def elect(self) -> str | None:
        alive = self._mesh.cluster_state().alive_nodes()
        if not alive:
            self._leader_id = None
            return None

        winner = max(alive, key=lambda n: n.node_id)
        self._leader_id = winner.node_id
        self._mesh.cluster_state().leader_id = winner.node_id

        # Update leader flag on all nodes
        for node in self._mesh.cluster_state().nodes.values():
            node.leader = node.node_id == winner.node_id

        logger.info("Leader elected: %s (%s)", winner.name, winner.node_id)
        return winner.node_id

    def on_node_joined(self, node_id: str) -> None:
        self.elect()

    def on_node_failed(self, node_id: str) -> None:
        if node_id == self._leader_id:
            logger.info("Leader %s failed, triggering election", node_id)
            self.elect()



"""ADK Runtime — stub for initial agent module import."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from astromesh_adk.agent import AgentWrapper, Agent

_default_runtime = None


def get_or_create_runtime():
    """Get or lazily create the default ADKRuntime."""
    global _default_runtime
    if _default_runtime is None:
        _default_runtime = ADKRuntime()
    return _default_runtime


class ADKRuntime:
    """Placeholder — full implementation in Task 13."""

    async def start(self) -> None:
        pass

    async def shutdown(self) -> None:
        pass

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.shutdown()
        return False

    async def run_agent(self, agent_wrapper, query, session_id, context, callbacks):
        raise NotImplementedError("ADKRuntime.run_agent not yet implemented")

    async def stream_agent(self, agent_wrapper, query, session_id, context, stream_steps, callbacks):
        raise NotImplementedError("ADKRuntime.stream_agent not yet implemented")
        yield  # make it a generator

    async def run_class_agent(self, agent_instance, query, session_id, context, callbacks):
        raise NotImplementedError("ADKRuntime.run_class_agent not yet implemented")

    async def stream_class_agent(self, agent_instance, query, session_id, context, stream_steps, callbacks):
        raise NotImplementedError("ADKRuntime.stream_class_agent not yet implemented")
        yield  # make it a generator

    async def run_team(self, team, query, session_id, context, callbacks):
        raise NotImplementedError("ADKRuntime.run_team not yet implemented")



"""Tests for RAG wrapper tools: rag_query, rag_ingest."""

from unittest.mock import AsyncMock, MagicMock
from astromesh.tools.base import ToolContext


def _ctx(**kwargs):
    return ToolContext(agent_name="test", session_id="s1", **kwargs)


class TestRagQueryTool:
    async def test_query_without_pipeline(self):
        from astromesh.tools.builtin.rag import RagQueryTool

        tool = RagQueryTool()
        result = await tool.execute({"query": "test"}, _ctx())
        assert result.success is False
        assert "pipeline" in result.error.lower()

    async def test_query_with_pipeline(self):
        from astromesh.tools.builtin.rag import RagQueryTool

        mock_pipeline = MagicMock()
        mock_pipeline.query = AsyncMock(return_value=[{"text": "Result doc", "score": 0.9}])
        ctx = _ctx()
        ctx.rag_pipeline = mock_pipeline

        tool = RagQueryTool()
        result = await tool.execute({"query": "find something", "top_k": 3}, ctx)
        assert result.success is True
        assert result.data["results"] == [{"text": "Result doc", "score": 0.9}]
        mock_pipeline.query.assert_called_once_with("find something", top_k=3)

    async def test_query_default_top_k(self):
        from astromesh.tools.builtin.rag import RagQueryTool

        mock_pipeline = MagicMock()
        mock_pipeline.query = AsyncMock(return_value=[])
        ctx = _ctx()
        ctx.rag_pipeline = mock_pipeline

        tool = RagQueryTool()
        result = await tool.execute({"query": "search"}, ctx)
        assert result.success is True
        mock_pipeline.query.assert_called_once_with("search", top_k=5)

    async def test_query_pipeline_exception(self):
        from astromesh.tools.builtin.rag import RagQueryTool

        mock_pipeline = MagicMock()
        mock_pipeline.query = AsyncMock(side_effect=RuntimeError("pipeline error"))
        ctx = _ctx()
        ctx.rag_pipeline = mock_pipeline

        tool = RagQueryTool()
        result = await tool.execute({"query": "bad query"}, ctx)
        assert result.success is False
        assert "pipeline error" in result.error


class TestRagIngestTool:
    async def test_ingest_without_pipeline(self):
        from astromesh.tools.builtin.rag import RagIngestTool

        tool = RagIngestTool()
        result = await tool.execute({"document": "content", "metadata": {}}, _ctx())
        assert result.success is False
        assert "pipeline" in result.error.lower()

    async def test_ingest_with_pipeline(self):
        from astromesh.tools.builtin.rag import RagIngestTool

        mock_pipeline = MagicMock()
        mock_pipeline.ingest = AsyncMock(return_value=None)
        ctx = _ctx()
        ctx.rag_pipeline = mock_pipeline

        tool = RagIngestTool()
        result = await tool.execute(
            {"document": "doc content", "metadata": {"source": "test"}}, ctx
        )
        assert result.success is True
        assert result.data["ingested"] is True
        mock_pipeline.ingest.assert_called_once_with("doc content", metadata={"source": "test"})

    async def test_ingest_default_empty_metadata(self):
        from astromesh.tools.builtin.rag import RagIngestTool

        mock_pipeline = MagicMock()
        mock_pipeline.ingest = AsyncMock(return_value=None)
        ctx = _ctx()
        ctx.rag_pipeline = mock_pipeline

        tool = RagIngestTool()
        result = await tool.execute({"document": "some text"}, ctx)
        assert result.success is True
        mock_pipeline.ingest.assert_called_once_with("some text", metadata={})

    async def test_ingest_pipeline_exception(self):
        from astromesh.tools.builtin.rag import RagIngestTool

        mock_pipeline = MagicMock()
        mock_pipeline.ingest = AsyncMock(side_effect=ValueError("ingest failed"))
        ctx = _ctx()
        ctx.rag_pipeline = mock_pipeline

        tool = RagIngestTool()
        result = await tool.execute({"document": "text"}, ctx)
        assert result.success is False
        assert "ingest failed" in result.error



import json
from io import StringIO

from astromesh.observability.tracing import TracingContext, Span, SpanStatus


class TestSpan:
    def test_span_creation(self):
        span = Span(name="test.span", trace_id="trace-1")
        assert span.name == "test.span"
        assert span.trace_id == "trace-1"
        assert span.parent_span_id is None
        assert span.status == SpanStatus.UNSET

    def test_span_finish(self):
        span = Span(name="test.span", trace_id="trace-1")
        span.finish(status=SpanStatus.OK)
        assert span.status == SpanStatus.OK
        assert span.duration_ms >= 0

    def test_span_set_attribute(self):
        span = Span(name="test.span", trace_id="trace-1")
        span.set_attribute("key", "value")
        assert span.attributes["key"] == "value"

    def test_span_add_event(self):
        span = Span(name="test.span", trace_id="trace-1")
        span.add_event("something.happened", {"detail": "info"})
        assert len(span.events) == 1
        assert span.events[0]["name"] == "something.happened"

    def test_span_to_dict(self):
        span = Span(name="test.span", trace_id="trace-1")
        span.set_attribute("key", "val")
        span.finish(status=SpanStatus.OK)
        d = span.to_dict()
        assert d["name"] == "test.span"
        assert d["trace_id"] == "trace-1"
        assert d["status"] == "ok"
        assert "duration_ms" in d


class TestTracingContext:
    def test_create_trace(self):
        ctx = TracingContext(agent_name="test-agent", session_id="sess-1")
        assert ctx.trace_id is not None
        assert ctx.agent_name == "test-agent"

    def test_start_and_finish_span(self):
        ctx = TracingContext(agent_name="test-agent", session_id="sess-1")
        span = ctx.start_span("test.operation")
        assert span.trace_id == ctx.trace_id
        span.finish(status=SpanStatus.OK)
        assert len(ctx.spans) == 1

    def test_nested_spans(self):
        ctx = TracingContext(agent_name="test-agent", session_id="sess-1")
        parent = ctx.start_span("parent")
        child = ctx.start_span("child")
        assert child.parent_span_id == parent.span_id
        child.finish(status=SpanStatus.OK)
        parent.finish(status=SpanStatus.OK)
        assert len(ctx.spans) == 2

    def test_to_dict(self):
        ctx = TracingContext(agent_name="test-agent", session_id="sess-1")
        span = ctx.start_span("op")
        span.finish(status=SpanStatus.OK)
        d = ctx.to_dict()
        assert d["trace_id"] == ctx.trace_id
        assert d["agent"] == "test-agent"
        assert len(d["spans"]) == 1

    def test_sample_rate_zero_disables_spans(self):
        ctx = TracingContext(agent_name="test-agent", session_id="sess-1", sample_rate=0.0)
        span = ctx.start_span("op")
        span.finish(status=SpanStatus.OK)
        assert ctx.is_sampled is False


class TestStructuredLogger:
    def test_log_event(self):
        from astromesh.observability.logging import StructuredLogger

        output = StringIO()
        logger = StructuredLogger(stream=output)
        logger.info(
            "tool.executed",
            agent="test-agent",
            trace_id="trace-1",
            tool="web_search",
            duration_ms=150,
            status="success",
        )
        line = output.getvalue().strip()
        data = json.loads(line)
        assert data["level"] == "info"
        assert data["event"] == "tool.executed"
        assert data["agent"] == "test-agent"
        assert data["tool"] == "web_search"
        assert "timestamp" in data

    def test_log_error(self):
        from astromesh.observability.logging import StructuredLogger

        output = StringIO()
        logger = StructuredLogger(stream=output)
        logger.error("tool.failed", tool="sql_query", error="Connection refused")
        line = output.getvalue().strip()
        data = json.loads(line)
        assert data["level"] == "error"
        assert data["error"] == "Connection refused"


class TestStdoutCollector:
    async def test_emit_trace(self):
        from astromesh.observability.collector import StdoutCollector

        output = StringIO()
        collector = StdoutCollector(stream=output)
        ctx = TracingContext(agent_name="test", session_id="s1")
        span = ctx.start_span("op")
        span.finish()
        await collector.emit_trace(ctx)
        line = output.getvalue().strip()
        data = json.loads(line)
        assert data["trace_id"] == ctx.trace_id


class TestInternalCollector:
    async def test_store_and_query_trace(self):
        from astromesh.observability.collector import InternalCollector

        collector = InternalCollector()
        ctx = TracingContext(agent_name="test-agent", session_id="s1")
        span = ctx.start_span("op")
        span.finish()
        await collector.emit_trace(ctx)
        traces = await collector.query_traces(agent="test-agent", limit=10)
        assert len(traces) == 1
        assert traces[0]["trace_id"] == ctx.trace_id

    async def test_query_by_trace_id(self):
        from astromesh.observability.collector import InternalCollector

        collector = InternalCollector()
        ctx = TracingContext(agent_name="test", session_id="s1")
        span = ctx.start_span("op")
        span.finish()
        await collector.emit_trace(ctx)
        trace = await collector.get_trace(ctx.trace_id)
        assert trace is not None
        assert trace["agent"] == "test"

    async def test_query_empty(self):
        from astromesh.observability.collector import InternalCollector

        collector = InternalCollector()
        traces = await collector.query_traces(agent="none", limit=10)
        assert traces == []



import os
from dataclasses import dataclass, field
from datetime import datetime

try:
    from astromesh._native import RustCostIndex

    _HAS_NATIVE_COST = True
except ImportError:
    _HAS_NATIVE_COST = False


@dataclass
class UsageRecord:
    agent_name: str
    session_id: str
    model: str
    provider: str
    input_tokens: int
    output_tokens: int
    cost_usd: float
    latency_ms: float
    pattern: str
    timestamp: datetime = field(default_factory=datetime.utcnow)


class CostTracker:
    """Track costs per agent, session, and model."""

    def __init__(self):
        self._records: list[UsageRecord] = []
        self._budgets: dict[str, float] = {}  # agent_name -> max_usd
        self._native_index = RustCostIndex() if _HAS_NATIVE_COST else None

    def record(self, record: UsageRecord):
        self._records.append(record)
        if self._native_index is not None and not os.environ.get("ASTROMESH_FORCE_PYTHON"):
            self._native_index.record(
                record.agent_name,
                record.session_id,
                record.model,
                record.provider,
                record.cost_usd,
                record.latency_ms,
                record.input_tokens,
                record.output_tokens,
                record.timestamp.timestamp(),
            )

    def set_budget(self, agent_name: str, max_usd: float):
        self._budgets[agent_name] = max_usd

    def check_budget(self, agent_name: str) -> dict:
        budget = self._budgets.get(agent_name)
        spent = self.get_total_cost(agent_name=agent_name)
        if budget is None:
            return {"has_budget": False, "spent": spent}
        return {
            "has_budget": True,
            "budget": budget,
            "spent": spent,
            "remaining": budget - spent,
            "exceeded": spent >= budget,
        }

    def get_total_cost(
        self,
        agent_name: str | None = None,
        session_id: str | None = None,
        since: datetime | None = None,
    ) -> float:
        if self._native_index is not None and not os.environ.get("ASTROMESH_FORCE_PYTHON"):
            return self._native_index.total_cost(
                agent_name,
                session_id,
                since.timestamp() if since else None,
            )
        records = self._records
        if agent_name:
            records = [r for r in records if r.agent_name == agent_name]
        if session_id:
            records = [r for r in records if r.session_id == session_id]
        if since:
            records = [r for r in records if r.timestamp >= since]
        return sum(r.cost_usd for r in records)

    def get_usage_summary(self, agent_name: str | None = None) -> dict:
        records = self._records
        if agent_name:
            records = [r for r in records if r.agent_name == agent_name]

        if not records:
            return {"total_cost": 0.0, "total_tokens": 0, "num_calls": 0}

        return {
            "total_cost": sum(r.cost_usd for r in records),
            "total_input_tokens": sum(r.input_tokens for r in records),
            "total_output_tokens": sum(r.output_tokens for r in records),
            "total_tokens": sum(r.input_tokens + r.output_tokens for r in records),
            "num_calls": len(records),
            "avg_latency_ms": sum(r.latency_ms for r in records) / len(records),
            "by_provider": self._group_by(records, "provider"),
            "by_model": self._group_by(records, "model"),
        }

    def _group_by(self, records: list[UsageRecord], field: str) -> dict:
        groups: dict[str, dict] = {}
        for r in records:
            key = getattr(r, field)
            if key not in groups:
                groups[key] = {"cost": 0.0, "calls": 0, "tokens": 0}
            groups[key]["cost"] += r.cost_usd
            groups[key]["calls"] += 1
            groups[key]["tokens"] += r.input_tokens + r.output_tokens
        return groups



"""Tests for MacOSInstaller."""

from pathlib import Path

from astromesh_node.installer.base import InstallerProtocol
from astromesh_node.installer.macos import MacOSInstaller


def test_implements_protocol():
    assert isinstance(MacOSInstaller(), InstallerProtocol)


def test_config_dir():
    assert MacOSInstaller().config_dir() == Path("/Library/Application Support/Astromesh/config")


def test_data_dir():
    assert MacOSInstaller().data_dir() == Path("/Library/Application Support/Astromesh/data")


def test_log_dir():
    assert MacOSInstaller().log_dir() == Path("/Library/Logs/Astromesh")


def test_bin_dir():
    assert MacOSInstaller().bin_dir() == Path("/usr/local/opt/astromesh/venv")



"""Shared fixtures. httpx ASGITransport does not run ASGI lifespan unless wrapped."""

import pytest
from asgi_lifespan import LifespanManager
from httpx import ASGITransport, AsyncClient

from astromesh.api.main import app


@pytest.fixture(autouse=True)
def _disable_agent_yaml_persist(monkeypatch):
    """Avoid writing into the repo's config/agents during API tests."""
    monkeypatch.setenv("ASTROMESH_PERSIST_AGENTS", "0")


@pytest.fixture
def use_native(monkeypatch):
    """Prefer Rust native extensions when installed (parity tests in test_native_*.py)."""
    monkeypatch.delenv("ASTROMESH_FORCE_PYTHON", raising=False)


@pytest.fixture
async def client():
    async with LifespanManager(app):
        async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as c:
            yield c



import pytest
from astromesh.runtime.engine import AgentRuntime

SAMPLE_CONFIG = {
    "apiVersion": "astromesh/v1",
    "kind": "Agent",
    "metadata": {"name": "test-status", "version": "1.0.0", "namespace": "test"},
    "spec": {
        "identity": {"display_name": "Test", "description": "Test agent"},
        "model": {
            "primary": {
                "provider": "ollama",
                "model": "llama3.1:8b",
                "endpoint": "http://localhost:11434",
            },
            "routing": {"strategy": "cost_optimized"},
        },
        "prompts": {"system": "You are a test agent."},
        "orchestration": {"pattern": "react", "max_iterations": 5},
    },
}


@pytest.fixture
def runtime(tmp_path):
    config_dir = tmp_path / "config"
    config_dir.mkdir()
    (config_dir / "agents").mkdir()
    return AgentRuntime(config_dir=str(config_dir))


async def test_register_agent_starts_as_draft(runtime):
    await runtime.register_agent(SAMPLE_CONFIG)
    agents = runtime.list_agents()
    agent = next(a for a in agents if a["name"] == "test-status")
    assert agent["status"] == "draft"


async def test_deploy_agent_changes_status(runtime):
    await runtime.register_agent(SAMPLE_CONFIG)
    await runtime.deploy_agent("test-status")
    agents = runtime.list_agents()
    agent = next(a for a in agents if a["name"] == "test-status")
    assert agent["status"] == "deployed"


async def test_pause_agent_changes_status(runtime):
    await runtime.register_agent(SAMPLE_CONFIG)
    await runtime.deploy_agent("test-status")
    runtime.pause_agent("test-status")
    agents = runtime.list_agents()
    agent = next(a for a in agents if a["name"] == "test-status")
    assert agent["status"] == "paused"


async def test_deploy_nonexistent_raises(runtime):
    with pytest.raises(ValueError, match="not found"):
        await runtime.deploy_agent("nonexistent")


async def test_pause_nondeployed_raises(runtime):
    await runtime.register_agent(SAMPLE_CONFIG)
    with pytest.raises(ValueError, match="not deployed"):
        runtime.pause_agent("test-status")


async def test_update_agent_config(runtime):
    await runtime.register_agent(SAMPLE_CONFIG)
    updated = {**SAMPLE_CONFIG}
    updated["spec"] = {**SAMPLE_CONFIG["spec"], "prompts": {"system": "Updated prompt"}}
    await runtime.update_agent("test-status", updated)
    agents = runtime.list_agents()
    agent = next(a for a in agents if a["name"] == "test-status")
    assert agent["status"] == "draft"  # update resets to draft


async def test_unregister_cleans_all_dicts(runtime):
    await runtime.register_agent(SAMPLE_CONFIG)
    runtime.unregister_agent("test-status")
    agents = runtime.list_agents()
    assert not any(a["name"] == "test-status" for a in agents)



"""astromeshctl dev — Hot-reload development server."""

import typer
from rich.panel import Panel

from astromesh_cli.output import console


def _launch_uvicorn(
    host: str = "0.0.0.0",
    port: int = 8000,
    reload: bool = True,
    reload_dirs: list[str] | None = None,
) -> None:
    """Launch uvicorn with the given configuration."""
    import uvicorn

    uvicorn.run(
        "astromesh.api.main:app",
        host=host,
        port=port,
        reload=reload,
        reload_dirs=reload_dirs or ["astromesh", "config"],
    )


def dev_command(
    host: str = typer.Option("0.0.0.0", "--host", help="Bind host"),
    port: int = typer.Option(8000, "--port", help="Bind port"),
    config: str = typer.Option("./config", "--config", help="Config directory"),
    no_open: bool = typer.Option(False, "--no-open", help="Skip opening browser"),
) -> None:
    """Start the Astromesh dev server with hot-reload."""
    banner = (
        f"[bold cyan]Astromesh Dev Server[/bold cyan]\n\n"
        f"  Host:   {host}\n"
        f"  Port:   {port}\n"
        f"  Config: {config}\n"
        f"  Reload: enabled"
    )
    console.print(Panel(banner, title="astromesh dev", border_style="cyan"))

    _launch_uvicorn(host=host, port=port, reload=True, reload_dirs=["astromesh", "config"])



"""Tests for LaunchdManager (macOS launchd adapter)."""

import pytest
from unittest.mock import patch, AsyncMock, MagicMock

from astromesh_node.platform.base import ServiceManagerProtocol
from astromesh_node.platform.launchd import LaunchdManager


@pytest.fixture
def manager():
    return LaunchdManager()


def test_implements_protocol(manager):
    assert isinstance(manager, ServiceManagerProtocol)


async def test_notify_ready_is_noop(manager):
    await manager.notify_ready()  # launchd assumes ready


async def test_notify_stopping_is_noop(manager):
    await manager.notify_stopping()


async def test_install_service_calls_launchctl(manager):
    with patch("astromesh_node.platform.launchd.asyncio") as mock_asyncio:
        proc = AsyncMock()
        proc.communicate.return_value = (b"", b"")
        proc.returncode = 0
        mock_asyncio.create_subprocess_exec = AsyncMock(return_value=proc)
        with patch("astromesh_node.platform.launchd.shutil"):
            with patch("astromesh_node.platform.launchd.Path") as mock_path:
                mock_path.return_value.exists.return_value = True
                await manager.install_service("full")


async def test_service_status_parses_launchctl(manager):
    launchctl_output = b'{\n\t"PID" = 1234;\n\t"Label" = "com.astromesh.daemon";\n};\n'
    with patch("astromesh_node.platform.launchd.asyncio") as mock_asyncio:
        proc = AsyncMock()
        proc.communicate.return_value = (launchctl_output, b"")
        proc.returncode = 0
        mock_asyncio.create_subprocess_exec = AsyncMock(return_value=proc)
        status = await manager.service_status()
        assert status["mode"] == "launchd"


def test_register_reload_handler_sets_sighup(manager):
    callback = MagicMock()
    with patch("astromesh_node.platform.launchd.signal") as mock_signal:
        manager.register_reload_handler(callback)
        mock_signal.signal.assert_called_once()
        assert mock_signal.signal.call_args[0][0] == mock_signal.SIGHUP



"""Benchmarks for rate limiter."""
import pytest


@pytest.mark.benchmark
class TestRateLimiterBenchmark:
    @pytest.fixture(params=[100, 1000, 10000], ids=["100-calls", "1000-calls", "10000-calls"])
    def n_calls(self, request):
        return request.param

    def test_native(self, benchmark, n_calls):
        try:
            from astromesh._native import RustRateLimiter
        except ImportError:
            pytest.skip("Native module not available")
        rl = RustRateLimiter()
        def run():
            for i in range(n_calls):
                rl.check(f"tool_{i % 10}", 3600.0, n_calls)
        benchmark(run)



"""Tests for LinuxInstaller."""

from pathlib import Path

from astromesh_node.installer.base import InstallerProtocol
from astromesh_node.installer.linux import LinuxInstaller


def test_implements_protocol():
    installer = LinuxInstaller()
    assert isinstance(installer, InstallerProtocol)


def test_config_dir():
    assert LinuxInstaller().config_dir() == Path("/etc/astromesh")


def test_data_dir():
    assert LinuxInstaller().data_dir() == Path("/var/lib/astromesh")


def test_log_dir():
    assert LinuxInstaller().log_dir() == Path("/var/log/astromesh")


def test_bin_dir():
    assert LinuxInstaller().bin_dir() == Path("/opt/astromesh/venv")


async def test_verify_returns_list():
    installer = LinuxInstaller()
    problems = await installer.verify()
    assert isinstance(problems, list)



"""Tests for communication tools: send_webhook, send_slack, send_email."""

import respx
import httpx
from astromesh.tools.base import ToolContext


def _ctx(**kwargs):
    return ToolContext(agent_name="test", session_id="s1", **kwargs)


class TestSendWebhookTool:
    @respx.mock
    async def test_send_webhook(self):
        from astromesh.tools.builtin.communication import SendWebhookTool

        respx.post("https://hooks.example.com/trigger").mock(
            return_value=httpx.Response(200, json={"ok": True})
        )
        tool = SendWebhookTool()
        result = await tool.execute(
            {"url": "https://hooks.example.com/trigger", "payload": {"message": "Hello"}},
            _ctx(),
        )
        assert result.success is True
        assert result.data["status_code"] == 200

    @respx.mock
    async def test_send_webhook_with_custom_headers(self):
        from astromesh.tools.builtin.communication import SendWebhookTool

        respx.post("https://hooks.example.com/trigger").mock(
            return_value=httpx.Response(201, text="created")
        )
        tool = SendWebhookTool()
        result = await tool.execute(
            {
                "url": "https://hooks.example.com/trigger",
                "payload": {"event": "test"},
                "headers": {"X-Custom": "value"},
            },
            _ctx(),
        )
        assert result.success is True
        assert result.data["status_code"] == 201

    @respx.mock
    async def test_send_webhook_network_error(self):
        from astromesh.tools.builtin.communication import SendWebhookTool

        respx.post("https://hooks.example.com/fail").mock(side_effect=httpx.ConnectError("refused"))
        tool = SendWebhookTool()
        result = await tool.execute(
            {"url": "https://hooks.example.com/fail", "payload": {}},
            _ctx(),
        )
        assert result.success is False
        assert result.error is not None

    @respx.mock
    async def test_send_webhook_metadata_contains_url(self):
        from astromesh.tools.builtin.communication import SendWebhookTool

        url = "https://hooks.example.com/trigger"
        respx.post(url).mock(return_value=httpx.Response(200, text="ok"))
        tool = SendWebhookTool()
        result = await tool.execute({"url": url, "payload": {}}, _ctx())
        assert result.metadata["url"] == url


class TestSendSlackTool:
    @respx.mock
    async def test_send_slack_webhook(self):
        from astromesh.tools.builtin.communication import SendSlackTool

        respx.post("https://hooks.slack.com/services/T/B/X").mock(
            return_value=httpx.Response(200, text="ok")
        )
        tool = SendSlackTool(config={"webhook_url": "https://hooks.slack.com/services/T/B/X"})
        result = await tool.execute({"message": "Hello Slack!"}, _ctx())
        assert result.success is True

    @respx.mock
    async def test_send_slack_via_secret(self):
        from astromesh.tools.builtin.communication import SendSlackTool

        respx.post("https://hooks.slack.com/services/T/B/Y").mock(
            return_value=httpx.Response(200, text="ok")
        )
        tool = SendSlackTool()
        ctx = _ctx(secrets={"SLACK_WEBHOOK_URL": "https://hooks.slack.com/services/T/B/Y"})
        result = await tool.execute({"message": "Hello via secret!"}, ctx)
        assert result.success is True

    async def test_send_slack_no_webhook_configured(self):
        from astromesh.tools.builtin.communication import SendSlackTool

        tool = SendSlackTool(config={})
        result = await tool.execute({"message": "Hello"}, _ctx())
        assert result.success is False
        assert result.error is not None

    @respx.mock
    async def test_send_slack_network_error(self):
        from astromesh.tools.builtin.communication import SendSlackTool

        respx.post("https://hooks.slack.com/services/T/B/Z").mock(
            side_effect=httpx.ConnectError("refused")
        )
        tool = SendSlackTool(config={"webhook_url": "https://hooks.slack.com/services/T/B/Z"})
        result = await tool.execute({"message": "Hello"}, _ctx())
        assert result.success is False


class TestSendEmailTool:
    async def test_missing_smtp_config(self):
        from astromesh.tools.builtin.communication import SendEmailTool

        tool = SendEmailTool(config={})
        result = await tool.execute(
            {"to": "user@example.com", "subject": "Test", "body": "Hello"}, _ctx()
        )
        assert result.success is False
        assert "smtp" in result.error.lower()

    async def test_missing_smtp_config_no_secret(self):
        from astromesh.tools.builtin.communication import SendEmailTool

        tool = SendEmailTool()
        result = await tool.execute(
            {"to": "a@b.com", "subject": "Hi", "body": "Body"}, _ctx(secrets={})
        )
        assert result.success is False

    async def test_smtp_sends_message(self, monkeypatch):
        """Test that email is sent when SMTP config is provided."""
        from astromesh.tools.builtin.communication import SendEmailTool
        import asyncio

        sent = {}

        def fake_send():
            sent["called"] = True

        async def fake_to_thread(fn, *args, **kwargs):
            fn()

        monkeypatch.setattr(asyncio, "to_thread", fake_to_thread)

        from unittest.mock import MagicMock, patch

        mock_server = MagicMock()
        mock_server.__enter__ = MagicMock(return_value=mock_server)
        mock_server.__exit__ = MagicMock(return_value=False)

        with patch("smtplib.SMTP", return_value=mock_server):
            tool = SendEmailTool(
                config={
                    "smtp_host": "smtp.example.com",
                    "smtp_port": 587,
                    "smtp_user": "user@example.com",
                    "smtp_password": "secret",
                    "from_address": "user@example.com",
                }
            )
            result = await tool.execute(
                {"to": "recipient@example.com", "subject": "Test", "body": "Hello"},
                _ctx(),
            )
        assert result.success is True
        assert result.data["to"] == "recipient@example.com"
        mock_server.starttls.assert_called_once()
        mock_server.login.assert_called_once_with("user@example.com", "secret")
        mock_server.send_message.assert_called_once()



from astromesh_adk.guardrails import normalize_guardrails_config


def test_none_returns_empty():
    assert normalize_guardrails_config(None) == {"input": [], "output": []}


def test_dict_with_lists():
    config = normalize_guardrails_config({
        "input": ["pii_detection", "topic_filter"],
        "output": ["pii_detection"],
    })
    assert len(config["input"]) == 2
    assert config["input"][0]["type"] == "pii_detection"
    assert config["input"][0]["action"] == "redact"
    assert config["input"][1]["type"] == "topic_filter"
    assert config["input"][1]["action"] == "block"


def test_dict_with_full_config():
    config = normalize_guardrails_config({
        "input": [{"type": "pii_detection", "action": "warn"}],
        "output": [],
    })
    assert config["input"][0]["action"] == "warn"


def test_only_input():
    config = normalize_guardrails_config({"input": ["max_length"]})
    assert config["output"] == []



import pytest
from datetime import datetime
from unittest.mock import AsyncMock

from astromesh.core.memory import (
    ConversationTurn,
    SemanticMemory,
    EpisodicMemory,
    MemoryManager,
)


def test_conversation_turn():
    turn = ConversationTurn(
        role="user", content="Hello", timestamp=datetime(2026, 1, 1)
    )
    assert turn.role == "user"
    assert turn.content == "Hello"
    assert turn.metadata == {}
    assert turn.token_count == 0


def test_semantic_memory():
    mem = SemanticMemory(
        content="Some fact", embedding=[0.1, 0.2], metadata={"source": "doc"}
    )
    assert mem.content == "Some fact"
    assert mem.embedding == [0.1, 0.2]
    assert mem.similarity == 0.0
    assert mem.source == ""


def test_episodic_memory():
    ep = EpisodicMemory(
        event_type="tool_call",
        summary="Called search",
        context={"query": "test"},
        outcome={"status": "ok"},
        timestamp=datetime(2026, 1, 1),
    )
    assert ep.event_type == "tool_call"
    assert ep.importance_score == 0.5


@pytest.mark.asyncio
async def test_memory_manager_build_context():
    """MemoryManager.build_context with sliding_window strategy."""
    turns = [
        ConversationTurn(
            role="user",
            content="Hi",
            timestamp=datetime(2026, 1, 1),
            token_count=5,
        ),
        ConversationTurn(
            role="assistant",
            content="Hello!",
            timestamp=datetime(2026, 1, 1),
            token_count=8,
        ),
    ]

    conversation = AsyncMock()
    conversation.get_history = AsyncMock(return_value=turns)

    config = {"conversational": {"strategy": "sliding_window"}}
    manager = MemoryManager(
        agent_id="agent-1", config=config, conversation=conversation
    )

    context = await manager.build_context(session_id="sess-1", current_query="test")

    assert context["conversation"] == turns
    conversation.get_history.assert_called_once_with("sess-1")


@pytest.mark.asyncio
async def test_memory_manager_persist_turn():
    """MemoryManager.persist_turn saves turn and checks history length."""
    turn = ConversationTurn(
        role="user", content="Hi", timestamp=datetime(2026, 1, 1), token_count=5
    )

    conversation = AsyncMock()
    conversation.save_turn = AsyncMock()
    conversation.get_history = AsyncMock(return_value=[turn])

    config = {"conversational": {"max_turns": 50}}
    manager = MemoryManager(
        agent_id="agent-1", config=config, conversation=conversation
    )

    await manager.persist_turn(session_id="sess-1", turn=turn)

    conversation.save_turn.assert_called_once_with("sess-1", turn)
    conversation.get_history.assert_called_once_with("sess-1")



"""Tests for MeshManager."""

from __future__ import annotations

import time
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from astromesh.mesh.config import MeshConfig
from astromesh.mesh.manager import MeshManager
from astromesh.mesh.state import ClusterState, NodeLoad, NodeState
from astromesh.runtime.services import ServiceManager


@pytest.fixture
def service_manager():
    return ServiceManager({"api": True, "agents": True, "tools": True})


@pytest.fixture
def mesh_config():
    return MeshConfig(
        enabled=True,
        node_name="test-node",
        bind="0.0.0.0:8000",
        seeds=["http://seed:8000"],
    )


@pytest.fixture
def mesh_config_no_seeds():
    return MeshConfig(
        enabled=True,
        node_name="test-node",
        bind="0.0.0.0:8000",
        seeds=[],
    )


@pytest.fixture
def manager(mesh_config, service_manager):
    return MeshManager(mesh_config, service_manager)


def test_manager_init(manager):
    assert manager.node_id is not None
    assert manager.node_id in manager.cluster_state().nodes


def test_manager_local_node_state(manager, service_manager):
    state = manager.local_node_state()
    assert state.name == "test-node"
    assert state.services == service_manager.enabled_services()


def test_manager_cluster_state(manager):
    cluster = manager.cluster_state()
    assert isinstance(cluster, ClusterState)
    assert len(cluster.nodes) == 1


def test_manager_is_alive(manager):
    assert manager.is_alive(manager.node_id) is True
    assert manager.is_alive("nonexistent-node") is False


def test_manager_update_node(manager):
    remote = NodeState(
        node_id="remote-1",
        name="remote-node",
        url="http://remote:8000",
        services=["api"],
        agents=[],
        load=NodeLoad(),
        joined_at=time.time(),
        last_heartbeat=time.time(),
    )
    manager.update_node("remote-1", remote)
    assert "remote-1" in manager.cluster_state().nodes


def test_manager_check_timeouts(manager):
    old_heartbeat = time.time() - 20  # > failure_timeout (15)
    remote = NodeState(
        node_id="remote-1",
        name="remote-node",
        url="http://remote:8000",
        services=["api"],
        agents=[],
        load=NodeLoad(),
        joined_at=time.time(),
        last_heartbeat=old_heartbeat,
    )
    manager.update_node("remote-1", remote)
    changed = manager.check_timeouts()
    assert "remote-1" in changed
    assert manager.cluster_state().nodes["remote-1"].status == "suspect"


def test_manager_check_timeouts_dead(manager):
    old_heartbeat = time.time() - 60  # > dead_timeout (30)
    remote = NodeState(
        node_id="remote-1",
        name="remote-node",
        url="http://remote:8000",
        services=["api"],
        agents=[],
        load=NodeLoad(),
        joined_at=time.time(),
        last_heartbeat=old_heartbeat,
    )
    manager.update_node("remote-1", remote)
    changed = manager.check_timeouts()
    assert "remote-1" in changed
    assert manager.cluster_state().nodes["remote-1"].status == "dead"


def test_manager_get_gossip_targets(manager):
    for i in range(5):
        remote = NodeState(
            node_id=f"remote-{i}",
            name=f"remote-{i}",
            url=f"http://remote-{i}:8000",
            services=["api"],
            agents=[],
            load=NodeLoad(),
            joined_at=time.time(),
            last_heartbeat=time.time(),
        )
        manager.update_node(f"remote-{i}", remote)

    targets = manager.get_gossip_targets()
    assert len(targets) == 3  # gossip_fanout default
    for t in targets:
        assert t.node_id != manager.node_id


def test_manager_get_gossip_targets_fewer_than_fanout(manager):
    remote = NodeState(
        node_id="remote-0",
        name="remote-0",
        url="http://remote-0:8000",
        services=["api"],
        agents=[],
        load=NodeLoad(),
        joined_at=time.time(),
        last_heartbeat=time.time(),
    )
    manager.update_node("remote-0", remote)
    targets = manager.get_gossip_targets()
    assert len(targets) == 1


@pytest.mark.asyncio
async def test_manager_join_success(manager):
    seed_node = NodeState(
        node_id="seed-node",
        name="seed",
        url="http://seed:8000",
        services=["api"],
        agents=[],
        load=NodeLoad(),
        joined_at=time.time(),
        last_heartbeat=time.time(),
    )
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.raise_for_status = MagicMock()
    mock_response.json.return_value = {
        "nodes": [seed_node.to_dict()],
        "leader_id": "seed-node",
    }

    manager._http.post = AsyncMock(return_value=mock_response)
    await manager.join()

    assert "seed-node" in manager.cluster_state().nodes
    assert manager.cluster_state().leader_id == "seed-node"


@pytest.mark.asyncio
async def test_manager_join_no_seeds(mesh_config_no_seeds, service_manager):
    mgr = MeshManager(mesh_config_no_seeds, service_manager)
    await mgr.join()
    assert len(mgr.cluster_state().nodes) == 1


@pytest.mark.asyncio
async def test_manager_leave(manager):
    manager._http.post = AsyncMock()
    await manager.leave()
    assert manager._left is True


@patch("astromesh.mesh.manager.psutil")
def test_manager_update_load(mock_psutil, manager):
    mock_psutil.cpu_percent.return_value = 42.0
    mock_mem = MagicMock()
    mock_mem.percent = 65.0
    mock_psutil.virtual_memory.return_value = mock_mem

    manager.update_load(active_requests=3)

    load = manager.local_node_state().load
    assert load.cpu_percent == 42.0
    assert load.memory_percent == 65.0
    assert load.active_requests == 3



"""Communication built-in tools: send_webhook, send_slack, send_email."""

from __future__ import annotations

import asyncio
import smtplib
from email.mime.text import MIMEText

import httpx

from astromesh.tools.base import BuiltinTool, ToolContext, ToolResult


class SendWebhookTool(BuiltinTool):
    name = "send_webhook"
    description = "Send an HTTP POST to a webhook URL with a JSON payload"
    parameters = {
        "type": "object",
        "properties": {
            "url": {"type": "string"},
            "payload": {"description": "JSON payload"},
            "headers": {"type": "object"},
        },
        "required": ["url", "payload"],
    }

    async def execute(self, arguments: dict, context: ToolContext) -> ToolResult:
        try:
            async with httpx.AsyncClient(timeout=30) as client:
                resp = await client.post(
                    arguments["url"],
                    json=arguments["payload"],
                    headers=arguments.get("headers", {}),
                )
                return ToolResult(
                    success=True,
                    data={"status_code": resp.status_code, "response": resp.text},
                    metadata={"url": arguments["url"]},
                )
        except Exception as e:
            return ToolResult(success=False, data=None, error=str(e))


class SendSlackTool(BuiltinTool):
    name = "send_slack"
    description = "Send a message to Slack via webhook"
    parameters = {
        "type": "object",
        "properties": {
            "message": {"type": "string"},
            "channel": {"type": "string"},
        },
        "required": ["message"],
    }

    async def execute(self, arguments: dict, context: ToolContext) -> ToolResult:
        webhook_url = self.config.get("webhook_url") or context.secrets.get("SLACK_WEBHOOK_URL")
        if not webhook_url:
            return ToolResult(success=False, data=None, error="No webhook_url configured")
        try:
            async with httpx.AsyncClient(timeout=10) as client:
                resp = await client.post(webhook_url, json={"text": arguments["message"]})
                return ToolResult(
                    success=True,
                    data={"response": resp.text},
                    metadata={"method": "webhook"},
                )
        except Exception as e:
            return ToolResult(success=False, data=None, error=str(e))


class SendEmailTool(BuiltinTool):
    name = "send_email"
    description = "Send an email via SMTP"
    parameters = {
        "type": "object",
        "properties": {
            "to": {"type": "string"},
            "subject": {"type": "string"},
            "body": {"type": "string"},
        },
        "required": ["to", "subject", "body"],
    }

    async def execute(self, arguments: dict, context: ToolContext) -> ToolResult:
        host = self.config.get("smtp_host") or context.secrets.get("SMTP_HOST")
        if not host:
            return ToolResult(success=False, data=None, error="SMTP host not configured")
        port = self.config.get("smtp_port", 587)
        user = self.config.get("smtp_user") or context.secrets.get("SMTP_USER")
        password = self.config.get("smtp_password") or context.secrets.get("SMTP_PASSWORD")
        from_addr = self.config.get("from_address", user)

        msg = MIMEText(arguments["body"])
        msg["Subject"] = arguments["subject"]
        msg["From"] = from_addr
        msg["To"] = arguments["to"]

        try:

            def _send():
                with smtplib.SMTP(host, port) as server:
                    server.starttls()
                    if user and password:
                        server.login(user, password)
                    server.send_message(msg)

            await asyncio.to_thread(_send)
            return ToolResult(
                success=True,
                data={"to": arguments["to"], "subject": arguments["subject"]},
                metadata={},
            )
        except Exception as e:
            return ToolResult(success=False, data=None, error=str(e))



"""Ollama provider adapter for the Astromesh Agent Runtime."""

from __future__ import annotations

import json
import time
from typing import Any, AsyncIterator

import httpx

from .base import CompletionChunk, CompletionResponse


class OllamaProvider:
    """Provider adapter that talks to a local Ollama instance."""

    def __init__(self, config: dict[str, Any] | None = None) -> None:
        config = config or {}
        self.base_url: str = config.get("base_url", "http://localhost:11434")
        self.model: str = config.get("model", "llama3")
        self.timeout: float = config.get("timeout", 120.0)
        self._client: httpx.AsyncClient | None = None

    async def _get_client(self) -> httpx.AsyncClient:
        if self._client is None or self._client.is_closed:
            self._client = httpx.AsyncClient(
                base_url=self.base_url,
                timeout=self.timeout,
            )
        return self._client

    # ------------------------------------------------------------------
    # Tool conversion helper
    # ------------------------------------------------------------------

    @staticmethod
    def _convert_tools(tools: list[dict]) -> list[dict]:
        """Convert a generic tool list into Ollama's expected format."""
        converted: list[dict] = []
        for tool in tools:
            if "function" in tool:
                converted.append(
                    {
                        "type": "function",
                        "function": {
                            "name": tool["function"]["name"],
                            "description": tool["function"].get("description", ""),
                            "parameters": tool["function"].get("parameters", {}),
                        },
                    }
                )
            else:
                converted.append(tool)
        return converted

    # ------------------------------------------------------------------
    # Multimodal conversion helper
    # ------------------------------------------------------------------

    @staticmethod
    def _convert_multimodal_messages(messages: list[dict]) -> list[dict]:
        """Convert OpenAI-style multimodal messages to Ollama format.

        Ollama expects images as base64 strings in a top-level ``images``
        list on the message, rather than inline ``image_url`` content parts.
        """
        converted: list[dict] = []
        for msg in messages:
            content = msg.get("content")
            if not isinstance(content, list):
                converted.append(msg)
                continue
            text_parts: list[str] = []
            images: list[str] = []
            for part in content:
                if part.get("type") == "text":
                    text_parts.append(part["text"])
                elif part.get("type") == "image_url":
                    url = part["image_url"]["url"]
                    # Strip data URI prefix to get raw base64.
                    if url.startswith("data:"):
                        url = url.split(",", 1)[-1]
                    images.append(url)
            new_msg = {**msg, "content": " ".join(text_parts)}
            if images:
                new_msg["images"] = images
            converted.append(new_msg)
        return converted

    # ------------------------------------------------------------------
    # ProviderProtocol implementation
    # ------------------------------------------------------------------

    async def complete(self, messages: list[dict], **kwargs: Any) -> CompletionResponse:
        client = await self._get_client()
        model = kwargs.pop("model", self.model)
        messages = self._convert_multimodal_messages(messages)

        payload: dict[str, Any] = {
            "model": model,
            "messages": messages,
            "stream": False,
        }

        tools = kwargs.pop("tools", None)
        if tools:
            payload["tools"] = self._convert_tools(tools)

        payload.update(kwargs)

        start = time.perf_counter()
        resp = await client.post("/api/chat", json=payload)
        latency_ms = (time.perf_counter() - start) * 1000
        resp.raise_for_status()
        data = resp.json()

        content = data.get("message", {}).get("content", "")
        tool_calls = data.get("message", {}).get("tool_calls", [])
        input_tokens = data.get("prompt_eval_count", 0)
        output_tokens = data.get("eval_count", 0)

        return CompletionResponse(
            content=content,
            model=model,
            provider="ollama",
            usage={"input_tokens": input_tokens, "output_tokens": output_tokens},
            latency_ms=latency_ms,
            cost=0.0,
            tool_calls=tool_calls,
        )

    async def stream(self, messages: list[dict], **kwargs: Any) -> AsyncIterator[CompletionChunk]:
        client = await self._get_client()
        model = kwargs.pop("model", self.model)
        messages = self._convert_multimodal_messages(messages)

        payload: dict[str, Any] = {
            "model": model,
            "messages": messages,
            "stream": True,
        }
        payload.update(kwargs)

        async with client.stream("POST", "/api/chat", json=payload) as resp:
            resp.raise_for_status()
            async for line in resp.aiter_lines():
                if not line:
                    continue
                chunk_data = json.loads(line)
                done = chunk_data.get("done", False)
                usage = None
                if done:
                    usage = {
                        "input_tokens": chunk_data.get("prompt_eval_count", 0),
                        "output_tokens": chunk_data.get("eval_count", 0),
                    }
                yield CompletionChunk(
                    content=chunk_data.get("message", {}).get("content", ""),
                    model=model,
                    provider="ollama",
                    done=done,
                    usage=usage,
                )

    async def health_check(self) -> bool:
        try:
            client = await self._get_client()
            resp = await client.get("/api/tags")
            return resp.status_code == 200
        except Exception:
            return False

    def supports_tools(self) -> bool:
        return True

    def supports_vision(self) -> bool:
        model_lower = self.model.lower()
        return "llava" in model_lower or "vision" in model_lower

    def estimated_cost(self, model: str, input_tokens: int, output_tokens: int) -> float:
        return 0.0



"""astromeshctl doctor command."""

import subprocess
import sys

import typer
from rich.table import Table

from astromesh_cli.client import api_get
from astromesh_cli.output import console, print_error, print_json

app = typer.Typer()

STATUS_ICONS = {
    "ok": "[green]OK[/green]",
    "degraded": "[yellow]DEGRADED[/yellow]",
    "error": "[red]ERROR[/red]",
    "unavailable": "[red]UNAVAILABLE[/red]",
}


def _check_stale_astromesh_package() -> str | None:
    """Return a warning string if the old 'astromesh' deb package is still installed."""
    if sys.platform != "linux":
        return None
    try:
        result = subprocess.run(
            ["dpkg", "-l", "astromesh"],
            capture_output=True,
            text=True,
            timeout=5,
        )
        if result.returncode == 0 and "astromesh" in result.stdout:
            # Ensure it's the old package, not the new astromesh-node package
            if "astromesh-node" not in result.stdout:
                return (
                    "Stale 'astromesh' package detected. "
                    "Upgrade to astromesh-node: sudo dpkg -i astromesh-node-*.deb"
                )
    except (FileNotFoundError, subprocess.TimeoutExpired):
        pass  # dpkg not available or timed out
    return None


@app.callback(invoke_without_command=True)
def doctor(json: bool = typer.Option(False, "--json", help="Output as JSON")):
    """Run system health checks."""
    migration_warning = _check_stale_astromesh_package()
    if migration_warning:
        console.print(f"[yellow]WARNING:[/yellow] {migration_warning}\n")

    try:
        data = api_get("/v1/system/doctor")
        if json:
            print_json(data)
            return

        healthy = data["healthy"]
        header = "[green]System Healthy[/green]" if healthy else "[red]System Unhealthy[/red]"
        console.print(f"\n{header}\n")

        table = Table()
        table.add_column("Check", style="cyan")
        table.add_column("Status")
        table.add_column("Message", style="dim")

        for name, check in data["checks"].items():
            status_display = STATUS_ICONS.get(check["status"], check["status"])
            table.add_row(name, status_display, check.get("message", ""))

        console.print(table)
    except Exception:
        print_error("Daemon not reachable at configured URL.")
        raise typer.Exit(code=0)



from fastapi import APIRouter
from pydantic import BaseModel

router = APIRouter()


class ToolExecuteRequest(BaseModel):
    tool_name: str
    arguments: dict = {}


@router.get("/tools")
async def list_tools():
    return {"tools": []}


@router.get("/tools/builtin")
async def list_builtin_tools():
    """Return all registered builtin tools with metadata."""
    from astromesh.tools import ToolLoader

    loader = ToolLoader()
    loader.auto_discover()

    tools = []
    for name in loader.list_available():
        tool_cls = loader.get(name)
        tools.append({
            "name": tool_cls.name,
            "description": tool_cls.description,
            "parameters": tool_cls.parameters,
        })

    return {"tools": tools, "count": len(tools)}


@router.post("/tools/execute")
async def execute_tool(request: ToolExecuteRequest):
    return {"tool": request.tool_name, "result": {}, "status": "not_configured"}



"""OrbitProvider Protocol and result data types."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Protocol, runtime_checkable

if TYPE_CHECKING:
    from astromesh_orbit.config import OrbitConfig


@dataclass
class CheckResult:
    name: str
    passed: bool
    message: str
    remediation: str | None = None


@dataclass
class ValidationResult:
    ok: bool
    checks: list[CheckResult]


@dataclass
class PlanResult:
    resources_to_create: list[str]
    resources_to_update: list[str]
    resources_to_destroy: list[str]
    raw_output: str
    estimated_monthly_cost: float | None = None


@dataclass
class ApplyResult:
    success: bool
    outputs: dict[str, str]
    raw_output: str


@dataclass
class ProvisionResult:
    apply: ApplyResult
    env_file: Path
    endpoints: dict[str, str]


@dataclass
class ResourceStatus:
    name: str
    resource_type: str
    status: str  # "running", "stopped", "error", "not_found"
    url: str | None = None


@dataclass
class DeploymentStatus:
    resources: list[ResourceStatus]
    state_bucket: str
    last_applied: str | None = None


@runtime_checkable
class OrbitProvider(Protocol):
    name: str

    async def validate(self, config: OrbitConfig) -> ValidationResult: ...
    async def generate(self, config: OrbitConfig, output_dir: Path) -> list[Path]: ...
    async def provision(self, config: OrbitConfig, output_dir: Path) -> ProvisionResult: ...
    async def status(self, config: OrbitConfig) -> DeploymentStatus: ...
    async def destroy(self, config: OrbitConfig, output_dir: Path) -> None: ...
    async def eject(self, config: OrbitConfig, output_dir: Path) -> Path: ...



"""MCP tools integration — lazy discovery."""

from __future__ import annotations

from typing import Any, Iterator


class MCPToolSet:
    """Lazy MCP tool set descriptor.

    Created synchronously at import time. Actual MCP server connection
    and tool discovery happens on first agent run via discover().
    """

    def __init__(self, config: dict):
        self.config = config
        self.discovered = False
        self._tools: list[dict] = []
        self._client: Any = None

    async def discover(self) -> list[dict]:
        """Connect to MCP server and discover available tools.

        Called automatically by the ADK runner on first agent execution.
        """
        from astromesh.mcp.client import MCPClient

        self._client = MCPClient(self.config)
        await self._client.connect()
        mcp_tools_info = self._client.get_tools()
        self._tools = [
            {
                "name": t.name,
                "description": t.description,
                "parameters": t.parameters,
            }
            for t in mcp_tools_info
        ]
        self.discovered = True
        return self._tools

    async def cleanup(self) -> None:
        """Disconnect from MCP server."""
        if self._client:
            await self._client.disconnect()
            self._client = None

    def __iter__(self) -> Iterator:
        """Allow unpacking with * in tools list.

        Before discovery, yields self as a lazy marker.
        After discovery, yields individual tool dicts.
        """
        if self.discovered:
            yield from self._tools
        else:
            yield self

    def __repr__(self) -> str:
        transport = self.config.get("transport", "unknown")
        return f"<MCPToolSet transport={transport!r} discovered={self.discovered}>"


def mcp_tools(
    transport: str,
    command: str | None = None,
    args: list[str] | None = None,
    url: str | None = None,
    headers: dict | None = None,
    env: dict | None = None,
) -> MCPToolSet:
    """Create a lazy MCP tool set.

    Returns immediately (no async). Connection happens on first agent run.
    Use with * unpacking in agent tools list:

        github = mcp_tools(transport="stdio", command="npx", args=[...])

        @agent(tools=[search, *github])
        async def my_agent(ctx): ...
    """
    config: dict[str, Any] = {"transport": transport}
    if command is not None:
        config["command"] = command
    if args is not None:
        config["args"] = args
    if url is not None:
        config["url"] = url
    if headers is not None:
        config["headers"] = headers
    if env is not None:
        config["env"] = env

    return MCPToolSet(config)



from abc import ABC, abstractmethod
import asyncio as aio
from dataclasses import dataclass
import json as json_mod
import os

try:
    from astromesh._native import rust_json_loads as _native_json_loads
except ImportError:
    _native_json_loads = None


def _loads(text):
    if _native_json_loads is not None and not os.environ.get("ASTROMESH_FORCE_PYTHON"):
        return _native_json_loads(text)
    return json_mod.loads(text)


@dataclass
class AgentStep:
    thought: str | None = None
    action: str | None = None
    action_input: dict | None = None
    observation: str | None = None
    result: str | None = None


class OrchestrationPattern(ABC):
    @abstractmethod
    async def execute(
        self, query, context, model_fn, tool_fn, tools, max_iterations=10
    ) -> dict: ...


class ReActPattern(OrchestrationPattern):
    """Thought -> Action -> Observation loop."""

    async def execute(self, query, context, model_fn, tool_fn, tools, max_iterations=10):
        messages = [{"role": "user", "content": query}]
        steps: list[AgentStep] = []
        for _ in range(max_iterations):
            response = await model_fn(messages, tools)
            if response.tool_calls:
                for tc in response.tool_calls:
                    observation = await tool_fn(tc["name"], tc["arguments"])
                    steps.append(
                        AgentStep(
                            thought=response.content,
                            action=tc["name"],
                            action_input=tc["arguments"],
                            observation=str(observation),
                        )
                    )
                    messages.append(
                        {
                            "role": "assistant",
                            "content": response.content,
                            "tool_calls": [tc],
                        }
                    )
                    messages.append(
                        {
                            "role": "tool",
                            "content": str(observation),
                            "tool_call_id": tc["id"],
                        }
                    )
            else:
                steps.append(AgentStep(result=response.content))
                return {"answer": response.content, "steps": steps}
        return {"answer": "Max iterations reached", "steps": steps}


class PlanAndExecutePattern(OrchestrationPattern):
    """Create plan, then execute each step sequentially."""

    async def execute(self, query, context, model_fn, tool_fn, tools, max_iterations=10):
        # Step 1: Ask model to create a plan
        plan_prompt = f'Create a step-by-step plan to answer: {query}\nReturn JSON: {{"steps": [{{"step": 1, "description": "...", "tool": null, "depends_on": []}}]}}'
        plan_response = await model_fn([{"role": "user", "content": plan_prompt}], tools)

        try:
            plan = _loads(plan_response.content)
            steps_plan = plan.get("steps", [])
        except (json_mod.JSONDecodeError, KeyError, ValueError):
            steps_plan = [{"step": 1, "description": query, "tool": None}]

        # Step 2: Execute each step
        steps = []
        results = []
        for step_info in steps_plan:
            step_query = f"Execute step {step_info['step']}: {step_info['description']}\nPrevious results: {results}"
            step_response = await model_fn([{"role": "user", "content": step_query}], tools)

            if step_response.tool_calls:
                for tc in step_response.tool_calls:
                    obs = await tool_fn(tc["name"], tc["arguments"])
                    results.append({"step": step_info["step"], "result": str(obs)})
                    steps.append(
                        AgentStep(
                            thought=step_response.content,
                            action=tc["name"],
                            action_input=tc["arguments"],
                            observation=str(obs),
                        )
                    )
            else:
                results.append({"step": step_info["step"], "result": step_response.content})
                steps.append(AgentStep(result=step_response.content))

        # Step 3: Synthesize final answer
        synthesis_prompt = (
            f"Synthesize a final answer from these results: {results}\nOriginal question: {query}"
        )
        final = await model_fn([{"role": "user", "content": synthesis_prompt}], [])
        steps.append(AgentStep(result=final.content))

        return {"answer": final.content, "steps": steps, "plan": steps_plan}


class ParallelFanOutPattern(OrchestrationPattern):
    """Fan out subtasks in parallel, then aggregate."""

    async def execute(self, query, context, model_fn, tool_fn, tools, max_iterations=10):
        # Decompose into subtasks
        decompose_prompt = (
            f"Decompose this into 2-4 independent subtasks (JSON list of strings): {query}"
        )
        decompose_resp = await model_fn([{"role": "user", "content": decompose_prompt}], [])

        try:
            subtasks = _loads(decompose_resp.content)
            if not isinstance(subtasks, list):
                subtasks = [query]
        except (json_mod.JSONDecodeError, Exception):
            subtasks = [query]

        # Execute subtasks in parallel
        async def run_subtask(subtask):
            resp = await model_fn([{"role": "user", "content": subtask}], tools)
            return {"subtask": subtask, "result": resp.content}

        results = await aio.gather(*[run_subtask(st) for st in subtasks])

        # Aggregate
        agg_prompt = f"Aggregate these results into a final answer:\n{json_mod.dumps(list(results))}\nOriginal question: {query}"
        final = await model_fn([{"role": "user", "content": agg_prompt}], [])

        steps = [AgentStep(thought=f"Subtask: {r['subtask']}", result=r["result"]) for r in results]
        steps.append(AgentStep(result=final.content))

        return {"answer": final.content, "steps": steps, "subtasks": list(results)}


class PipelinePattern(OrchestrationPattern):
    """Sequential pipeline: output of step N feeds into step N+1."""

    def __init__(self, stages: list[str] | None = None):
        self._stages = stages or ["analyze", "process", "synthesize"]

    async def execute(self, query, context, model_fn, tool_fn, tools, max_iterations=10):
        current_input = query
        steps = []

        for stage in self._stages:
            prompt = f"Stage '{stage}': Process the following input and produce output for the next stage.\nInput: {current_input}"
            response = await model_fn([{"role": "user", "content": prompt}], tools)

            if response.tool_calls:
                for tc in response.tool_calls:
                    obs = await tool_fn(tc["name"], tc["arguments"])
                    steps.append(
                        AgentStep(
                            thought=f"Stage: {stage}",
                            action=tc["name"],
                            action_input=tc["arguments"],
                            observation=str(obs),
                        )
                    )
                    current_input = str(obs)
            else:
                steps.append(AgentStep(thought=f"Stage: {stage}", result=response.content))
                current_input = response.content

        return {"answer": current_input, "steps": steps}



"""Tests for daemon config loading and path detection."""

from unittest.mock import patch

from astromesh_node.daemon.config import DaemonConfig, detect_config_dir


def test_daemon_config_defaults():
    cfg = DaemonConfig()
    assert cfg.host == "0.0.0.0"
    assert cfg.port == 8000
    assert cfg.log_level == "info"


def test_daemon_config_from_dict():
    data = {
        "spec": {
            "api": {"host": "127.0.0.1", "port": 9000},
            "services": {"api": True, "agents": True},
            "peers": [{"name": "node2"}],
            "mesh": {"enabled": True},
        }
    }
    cfg = DaemonConfig.from_dict(data)
    assert cfg.host == "127.0.0.1"
    assert cfg.port == 9000
    assert cfg.services == {"api": True, "agents": True}


def test_detect_config_dir_explicit():
    assert detect_config_dir("/custom/path") == "/custom/path"


def test_detect_config_dir_dev_mode(tmp_path):
    config_dir = tmp_path / "config"
    config_dir.mkdir()
    (config_dir / "runtime.yaml").write_text("spec: {}")
    with patch("astromesh_node.daemon.config.Path.cwd", return_value=tmp_path):
        result = detect_config_dir(None)
        assert result == str(config_dir)



import uuid

from qdrant_client import QdrantClient
from qdrant_client.models import Distance, PointStruct, VectorParams

from astromesh.core.memory import SemanticBackend, SemanticMemory


class QdrantSemanticBackend(SemanticBackend):
    def __init__(
        self,
        client: QdrantClient | None = None,
        url: str | None = None,
        dimension: int = 1536,
    ):
        if client is not None:
            self._client = client
        elif url:
            self._client = QdrantClient(url=url)
        else:
            self._client = QdrantClient(":memory:")
        self._dimension = dimension

    def _collection_name(self, agent_id: str) -> str:
        return f"agent_{agent_id}"

    def _ensure_collection(self, agent_id: str):
        name = self._collection_name(agent_id)
        collections = [c.name for c in self._client.get_collections().collections]
        if name not in collections:
            self._client.create_collection(
                collection_name=name,
                vectors_config=VectorParams(
                    size=self._dimension,
                    distance=Distance.COSINE,
                ),
            )

    async def store(self, agent_id, content, embedding, metadata):
        self._ensure_collection(agent_id)
        memory_id = str(uuid.uuid4())
        payload = {"content": content, "metadata": metadata}
        self._client.upsert(
            collection_name=self._collection_name(agent_id),
            points=[
                PointStruct(
                    id=memory_id,
                    vector=embedding,
                    payload=payload,
                )
            ],
        )
        return memory_id

    async def search(self, agent_id, query_embedding, top_k=10, threshold=0.7):
        self._ensure_collection(agent_id)
        results = self._client.search(
            collection_name=self._collection_name(agent_id),
            query_vector=query_embedding,
            limit=top_k,
            score_threshold=threshold,
        )
        return [
            SemanticMemory(
                content=hit.payload.get("content", ""),
                embedding=[],
                metadata=hit.payload.get("metadata", {}),
                similarity=hit.score,
                source=str(hit.id),
            )
            for hit in results
        ]

    async def delete(self, agent_id, memory_id):
        self._client.delete(
            collection_name=self._collection_name(agent_id),
            points_selector=[memory_id],
        )



"""Tests for WindowsServiceManager (Windows Service adapter)."""

import pytest
from unittest.mock import patch, MagicMock, AsyncMock

from astromesh_node.platform.base import ServiceManagerProtocol
from astromesh_node.platform.windows import WindowsServiceManager


@pytest.fixture
def manager():
    return WindowsServiceManager()


def test_implements_protocol(manager):
    assert isinstance(manager, ServiceManagerProtocol)


async def test_notify_ready_is_noop_without_win32(manager):
    """notify_ready gracefully degrades when not running as Windows Service."""
    with patch.object(manager, "_service_handle", None):
        await manager.notify_ready()  # Should not raise


async def test_notify_stopping_is_noop_without_win32(manager):
    with patch.object(manager, "_service_handle", None):
        await manager.notify_stopping()


async def test_service_status_returns_dict(manager):
    with patch("astromesh_node.platform.windows.asyncio") as mock_asyncio:
        proc = AsyncMock()
        proc.communicate.return_value = (
            b"STATE              : 4  RUNNING\nPID                : 5678\n",
            b"",
        )
        proc.returncode = 0
        mock_asyncio.create_subprocess_exec = AsyncMock(return_value=proc)
        mock_asyncio.subprocess = MagicMock()
        mock_asyncio.subprocess.PIPE = -1
        status = await manager.service_status()
        assert status["mode"] == "windows_service"
        assert status["running"] is True


def test_register_reload_handler_stores_callback(manager):
    callback = MagicMock()
    manager.register_reload_handler(callback)
    assert manager._reload_handler == callback



"""ONNX Runtime provider adapter for the Astromesh Agent Runtime."""

from __future__ import annotations

import time
from typing import Any, AsyncIterator

from .base import CompletionChunk, CompletionResponse


class ONNXProvider:
    """Provider adapter that runs inference locally via ONNX Runtime.

    This provider is *not* HTTP-based.  It loads an ONNX model from disk and
    runs inference in-process.  The model is lazily loaded on first use.
    """

    def __init__(self, config: dict[str, Any] | None = None) -> None:
        config = config or {}
        self.model_path: str = config.get("model_path", "")
        self.model_name: str = config.get("model", "onnx-local")
        self._session: Any = None

    def _load_model(self) -> Any:
        """Lazily load the ONNX model into an InferenceSession."""
        if self._session is None:
            try:
                import onnxruntime as ort  # type: ignore[import-untyped]

                self._session = ort.InferenceSession(self.model_path)
            except ImportError as exc:
                raise ImportError(
                    "onnxruntime is required for ONNXProvider. "
                    "Install it with: pip install onnxruntime"
                ) from exc
        return self._session

    async def complete(self, messages: list[dict], **kwargs: Any) -> CompletionResponse:
        session = self._load_model()

        # Build a simple text prompt from messages
        prompt = "\n".join(m.get("content", "") for m in messages)

        start = time.perf_counter()
        # Run inference — the exact I/O depends on the model.
        input_name = session.get_inputs()[0].name
        import numpy as np  # type: ignore[import-untyped]

        input_ids = np.array([[ord(c) for c in prompt[:512]]], dtype=np.int64)
        outputs = session.run(None, {input_name: input_ids})
        latency_ms = (time.perf_counter() - start) * 1000

        content = str(outputs[0]) if outputs else ""

        return CompletionResponse(
            content=content,
            model=self.model_name,
            provider="onnx",
            usage={"input_tokens": len(prompt.split()), "output_tokens": len(content.split())},
            latency_ms=latency_ms,
            cost=0.0,
        )

    async def stream(self, messages: list[dict], **kwargs: Any) -> AsyncIterator[CompletionChunk]:
        result = await self.complete(messages, **kwargs)
        yield CompletionChunk(
            content=result.content,
            model=result.model,
            provider="onnx",
            done=True,
            usage=result.usage,
        )

    async def health_check(self) -> bool:
        try:
            self._load_model()
            return True
        except Exception:
            return False

    def supports_tools(self) -> bool:
        return False

    def supports_vision(self) -> bool:
        return False

    def estimated_cost(self, model: str, input_tokens: int, output_tokens: int) -> float:
        return 0.0



from __future__ import annotations

import json
import sys
from abc import ABC, abstractmethod
from collections import deque
from typing import IO, TYPE_CHECKING

from astromesh.observability.tracing import TracingContext

if TYPE_CHECKING:
    from astromesh.observability.telemetry import TelemetryManager


class Collector(ABC):
    @abstractmethod
    async def emit_trace(self, ctx: TracingContext) -> None: ...

    async def query_traces(self, agent: str | None = None, limit: int = 20) -> list[dict]:
        return []

    async def get_trace(self, trace_id: str) -> dict | None:
        return None


class StdoutCollector(Collector):
    def __init__(self, stream: IO | None = None):
        self._stream = stream or sys.stdout

    async def emit_trace(self, ctx: TracingContext) -> None:
        self._stream.write(json.dumps(ctx.to_dict(), default=str) + "\n")
        self._stream.flush()


class InternalCollector(Collector):
    """In-memory collector for development and small deployments."""

    def __init__(self, max_traces: int = 10000):
        self._traces: deque[dict] = deque(maxlen=max_traces)
        self._index: dict[str, dict] = {}

    async def emit_trace(self, ctx: TracingContext) -> None:
        trace_data = ctx.to_dict()
        self._traces.append(trace_data)
        self._index[ctx.trace_id] = trace_data

    async def query_traces(self, agent: str | None = None, limit: int = 20) -> list[dict]:
        results = list(self._traces)
        if agent:
            results = [t for t in results if t.get("agent") == agent]
        return list(reversed(results))[:limit]

    async def get_trace(self, trace_id: str) -> dict | None:
        return self._index.get(trace_id)


class OTLPCollector(InternalCollector):
    """Bridges TracingContext spans to OpenTelemetry via TelemetryManager."""

    def __init__(
        self,
        telemetry_manager: TelemetryManager | None = None,
        max_traces: int = 1000,
    ):
        super().__init__(max_traces=max_traces)
        self._telemetry = telemetry_manager

    async def emit_trace(self, ctx: TracingContext) -> None:
        await super().emit_trace(ctx)
        if self._telemetry and self._telemetry.get_tracer():
            tracer = self._telemetry.get_tracer()
            for span_data in ctx.to_dict().get("spans", []):
                with tracer.start_as_current_span(span_data["name"]) as otel_span:
                    for k, v in span_data.get("attributes", {}).items():
                        otel_span.set_attribute(k, v)



# tests/test_workflow_loader.py
import pytest
from astromesh.workflow.loader import WorkflowLoader
from astromesh.workflow.models import WorkflowSpec, StepType


@pytest.fixture
def workflows_dir(tmp_path):
    wf_dir = tmp_path / "workflows"
    wf_dir.mkdir()
    return wf_dir


def _write_workflow(wf_dir, name, content):
    (wf_dir / f"{name}.workflow.yaml").write_text(content)


class TestWorkflowLoader:
    def test_load_empty_dir(self, workflows_dir):
        loader = WorkflowLoader(str(workflows_dir))
        workflows = loader.load_all()
        assert workflows == {}

    def test_load_basic_workflow(self, workflows_dir):
        _write_workflow(workflows_dir, "test", """
apiVersion: astromesh/v1
kind: Workflow
metadata:
  name: test-wf
  version: "0.1.0"
spec:
  trigger: api
  timeout_seconds: 120
  steps:
    - name: greet
      agent: greeter
      input: "{{ trigger.query }}"
""")
        loader = WorkflowLoader(str(workflows_dir))
        workflows = loader.load_all()
        assert "test-wf" in workflows
        wf = workflows["test-wf"]
        assert isinstance(wf, WorkflowSpec)
        assert wf.trigger == "api"
        assert wf.timeout_seconds == 120
        assert len(wf.steps) == 1
        assert wf.steps[0].step_type == StepType.AGENT

    def test_load_workflow_with_tool_step(self, workflows_dir):
        _write_workflow(workflows_dir, "tool-wf", """
apiVersion: astromesh/v1
kind: Workflow
metadata:
  name: tool-wf
spec:
  steps:
    - name: store
      tool: cache_store
      arguments:
        key: "test-key"
        value: "test-value"
""")
        loader = WorkflowLoader(str(workflows_dir))
        workflows = loader.load_all()
        wf = workflows["tool-wf"]
        assert wf.steps[0].step_type == StepType.TOOL
        assert wf.steps[0].arguments["key"] == "test-key"

    def test_load_workflow_with_switch_step(self, workflows_dir):
        _write_workflow(workflows_dir, "switch-wf", """
apiVersion: astromesh/v1
kind: Workflow
metadata:
  name: switch-wf
spec:
  steps:
    - name: check
      agent: checker
      input: "{{ trigger.query }}"
    - name: decide
      switch:
        - when: "{{ steps.check.output.score > 5 }}"
          goto: high
        - default: true
          goto: low
    - name: high
      agent: high-handler
      input: "high path"
    - name: low
      agent: low-handler
      input: "low path"
""")
        loader = WorkflowLoader(str(workflows_dir))
        workflows = loader.load_all()
        wf = workflows["switch-wf"]
        assert wf.steps[1].step_type == StepType.SWITCH
        assert len(wf.steps[1].switch) == 2

    def test_load_workflow_with_retry(self, workflows_dir):
        _write_workflow(workflows_dir, "retry-wf", """
apiVersion: astromesh/v1
kind: Workflow
metadata:
  name: retry-wf
spec:
  steps:
    - name: flaky
      agent: flaky-agent
      input: "{{ trigger.query }}"
      retry:
        max_attempts: 3
        backoff: exponential
        initial_delay_seconds: 2
      timeout_seconds: 30
      on_error: fallback
    - name: fallback
      agent: fallback-agent
      input: "error occurred"
""")
        loader = WorkflowLoader(str(workflows_dir))
        workflows = loader.load_all()
        wf = workflows["retry-wf"]
        step = wf.steps[0]
        assert step.retry.max_attempts == 3
        assert step.retry.backoff == "exponential"
        assert step.timeout_seconds == 30
        assert step.on_error == "fallback"

    def test_invalid_kind_skipped(self, workflows_dir):
        _write_workflow(workflows_dir, "agent", """
apiVersion: astromesh/v1
kind: Agent
metadata:
  name: not-a-workflow
spec:
  identity:
    description: "This is an agent, not a workflow"
""")
        loader = WorkflowLoader(str(workflows_dir))
        workflows = loader.load_all()
        assert workflows == {}

    def test_missing_dir_returns_empty(self, tmp_path):
        loader = WorkflowLoader(str(tmp_path / "nonexistent"))
        workflows = loader.load_all()
        assert workflows == {}

    def test_load_single_workflow(self, workflows_dir):
        _write_workflow(workflows_dir, "single", """
apiVersion: astromesh/v1
kind: Workflow
metadata:
  name: single-wf
spec:
  steps:
    - name: do-thing
      agent: doer
      input: "{{ trigger.query }}"
""")
        loader = WorkflowLoader(str(workflows_dir))
        wf = loader.load_file(workflows_dir / "single.workflow.yaml")
        assert wf.name == "single-wf"



# tests/test_dashboard_api.py
from httpx import AsyncClient, ASGITransport
from fastapi import FastAPI

from astromesh.api.routes.dashboard import router


def _make_app():
    app = FastAPI()
    app.include_router(router, prefix="/v1")
    return app


class TestDashboardAPI:
    async def test_dashboard_returns_html(self):
        app = _make_app()
        async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as ac:
            resp = await ac.get("/v1/dashboard/")
        assert resp.status_code == 200
        assert "text/html" in resp.headers["content-type"]
        assert "<html" in resp.text.lower()

    async def test_dashboard_contains_key_elements(self):
        app = _make_app()
        async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as ac:
            resp = await ac.get("/v1/dashboard/")
        html = resp.text
        # Verify key UI sections exist
        assert "traces" in html.lower() or "Traces" in html
        assert "metrics" in html.lower() or "Metrics" in html
        assert "/v1/traces" in html  # fetches from traces API
        assert "/v1/metrics" in html  # fetches from metrics API

    async def test_dashboard_contains_workflow_section(self):
        app = _make_app()
        async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as ac:
            resp = await ac.get("/v1/dashboard/")
        html = resp.text
        assert "workflow" in html.lower()

    async def test_dashboard_has_auto_refresh(self):
        app = _make_app()
        async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as ac:
            resp = await ac.get("/v1/dashboard/")
        html = resp.text
        # Should have some form of auto-refresh (setInterval or similar)
        assert "setInterval" in html or "auto-refresh" in html.lower()



import json
import pytest
from dataclasses import dataclass
from unittest.mock import AsyncMock

from astromesh.orchestration.swarm import SwarmPattern


@dataclass
class CompletionResponse:
    content: str
    tool_calls: list | None = None


def make_response(content, tool_calls=None):
    return CompletionResponse(content=content, tool_calls=tool_calls)


class TestSwarmWithAgentTools:
    @pytest.mark.asyncio
    async def test_swarm_handoff_via_tool_fn(self):
        """Swarm handoff should invoke the target agent via tool_fn."""
        handoff_json = json.dumps(
            {"handoff": "specialist", "context": "Need expert analysis"}
        )
        model_fn = AsyncMock(
            side_effect=[
                make_response(handoff_json),
                make_response("Expert analysis complete."),
            ]
        )
        tool_fn = AsyncMock(
            return_value={"answer": "Specialist context loaded", "steps": []}
        )

        pattern = SwarmPattern(agent_configs={"specialist": "specialist-agent"})
        result = await pattern.execute(
            query="Analyze this data",
            context={},
            model_fn=model_fn,
            tool_fn=tool_fn,
            tools=[],
        )

        tool_fn.assert_called_once_with(
            "specialist", {"query": "Need expert analysis"}
        )
        assert result["answer"] == "Expert analysis complete."

    @pytest.mark.asyncio
    async def test_swarm_no_handoff_returns_directly(self):
        """If no handoff, swarm returns the agent's direct answer."""
        model_fn = AsyncMock(return_value=make_response("Direct answer."))
        tool_fn = AsyncMock()

        pattern = SwarmPattern()
        result = await pattern.execute(
            query="Simple question",
            context={},
            model_fn=model_fn,
            tool_fn=tool_fn,
            tools=[],
        )

        assert result["answer"] == "Direct answer."
        tool_fn.assert_not_called()

    @pytest.mark.asyncio
    async def test_swarm_multiple_handoffs(self):
        """Chain of handoffs: A -> B -> C -> final answer."""
        model_fn = AsyncMock(
            side_effect=[
                make_response(
                    json.dumps({"handoff": "agent-b", "context": "step 1"})
                ),
                make_response(
                    json.dumps({"handoff": "agent-c", "context": "step 2"})
                ),
                make_response("Final from agent-c."),
            ]
        )
        tool_fn = AsyncMock(
            side_effect=[
                {"answer": "B loaded", "steps": []},
                {"answer": "C loaded", "steps": []},
            ]
        )

        pattern = SwarmPattern(
            agent_configs={"agent-b": "agent-b", "agent-c": "agent-c"}
        )
        result = await pattern.execute(
            query="Multi-step task",
            context={},
            model_fn=model_fn,
            tool_fn=tool_fn,
            tools=[],
        )

        assert tool_fn.call_count == 2
        assert result["answer"] == "Final from agent-c."
        assert result["final_agent"] == "agent-c"

    @pytest.mark.asyncio
    async def test_swarm_tool_calls_still_work(self):
        """Regular tool calls (non-handoff) still work within swarm."""
        tool_call = {"id": "tc_1", "name": "web_search", "arguments": {"q": "test"}}
        model_fn = AsyncMock(
            side_effect=[
                make_response("Searching...", tool_calls=[tool_call]),
                make_response("Found the answer."),
            ]
        )
        tool_fn = AsyncMock(return_value={"results": ["result1"]})

        pattern = SwarmPattern()
        result = await pattern.execute(
            query="Search for something",
            context={},
            model_fn=model_fn,
            tool_fn=tool_fn,
            tools=[],
        )

        assert result["answer"] == "Found the answer."



import hashlib
import hmac
import httpx
import logging
import os

from astromesh.channels.base import ChannelAdapter, ChannelMessage, MediaAttachment

logger = logging.getLogger(__name__)

# WhatsApp media type to generic media type mapping.
_WA_MEDIA_TYPES = {
    "image": "image",
    "audio": "audio",
    "video": "video",
    "document": "document",
}


class WhatsAppClient(ChannelAdapter):
    GRAPH_API_URL = "https://graph.facebook.com/v21.0"

    def __init__(self):
        self.verify_token = os.environ.get("WHATSAPP_VERIFY_TOKEN", "")
        self.access_token = os.environ.get("WHATSAPP_ACCESS_TOKEN", "")
        self.phone_number_id = os.environ.get("WHATSAPP_PHONE_NUMBER_ID", "")
        self.app_secret = os.environ.get("WHATSAPP_APP_SECRET", "")

    def verify_webhook(self, mode: str, token: str, challenge: str) -> str | None:
        """Verify Meta webhook subscription. Returns challenge if valid, None otherwise."""
        if mode == "subscribe" and token == self.verify_token:
            return challenge
        return None

    def verify_request(self, payload: bytes, signature: str) -> bool:
        """Validate X-Hub-Signature-256 header from Meta."""
        if not self.app_secret:
            return True  # Skip validation if no secret configured
        expected = (
            "sha256=" + hmac.new(self.app_secret.encode(), payload, hashlib.sha256).hexdigest()
        )
        return hmac.compare_digest(expected, signature)

    async def parse_incoming(self, payload: dict) -> list[ChannelMessage]:
        """Parse Meta webhook payload and extract text and media messages."""
        messages: list[ChannelMessage] = []
        for entry in payload.get("entry", []):
            for change in entry.get("changes", []):
                value = change.get("value", {})
                for msg in value.get("messages", []):
                    msg_type = msg.get("type", "")
                    text: str | None = None
                    media: list[MediaAttachment] = []

                    if msg_type == "text":
                        text = msg["text"]["body"]
                    elif msg_type in _WA_MEDIA_TYPES:
                        media_data = msg.get(msg_type, {})
                        media.append(
                            MediaAttachment(
                                media_type=_WA_MEDIA_TYPES[msg_type],
                                mime_type=media_data.get("mime_type", f"{msg_type}/*"),
                                content=None,
                                source_id=media_data.get("id", ""),
                                filename=media_data.get("filename"),
                            )
                        )
                        # Some media messages include a caption as text.
                        caption = media_data.get("caption")
                        if caption:
                            text = caption
                    else:
                        continue  # Skip unsupported message types.

                    messages.append(
                        ChannelMessage(
                            sender_id=msg["from"],
                            text=text,
                            media=media,
                            message_id=msg["id"],
                            timestamp=msg.get("timestamp", ""),
                            channel="whatsapp",
                            raw_payload=msg,
                        )
                    )
        return messages

    async def send_text(self, recipient_id: str, text: str) -> dict:
        """Send a text message via WhatsApp Cloud API."""
        url = f"{self.GRAPH_API_URL}/{self.phone_number_id}/messages"
        headers = {
            "Authorization": f"Bearer {self.access_token}",
            "Content-Type": "application/json",
        }
        data = {
            "messaging_product": "whatsapp",
            "to": recipient_id,
            "type": "text",
            "text": {"body": text},
        }
        async with httpx.AsyncClient() as client:
            resp = await client.post(url, json=data, headers=headers)
            resp.raise_for_status()
            return resp.json()

    async def download_media(self, attachment: MediaAttachment) -> bytes:
        """Download media from WhatsApp via the Graph API.

        Two-step process: first retrieve the download URL, then fetch the bytes.
        Media URLs expire after ~5 minutes.
        """
        headers = {"Authorization": f"Bearer {self.access_token}"}
        async with httpx.AsyncClient() as client:
            # Step 1: Get the download URL.
            meta_resp = await client.get(
                f"{self.GRAPH_API_URL}/{attachment.source_id}",
                headers=headers,
            )
            meta_resp.raise_for_status()
            download_url = meta_resp.json()["url"]

            # Step 2: Download the actual media bytes.
            media_resp = await client.get(download_url, headers=headers)
            media_resp.raise_for_status()
            return media_resp.content



"""Tests for native cost tracker."""
import pytest


class TestCostIndex:
    def test_record_and_total(self):
        try:
            from astromesh._native import RustCostIndex
            idx = RustCostIndex()
            idx.record("agent1", "s1", "gpt-4", "openai", 0.05, 100.0, 500, 200, 1000.0)
            idx.record("agent1", "s1", "gpt-4", "openai", 0.03, 80.0, 300, 100, 1001.0)
            assert abs(idx.total_cost(None, None, None) - 0.08) < 0.001
        except ImportError:
            pytest.skip("Native module not available")

    def test_filter_by_agent(self):
        try:
            from astromesh._native import RustCostIndex
            idx = RustCostIndex()
            idx.record("agent1", "s1", "gpt-4", "openai", 0.05, 100.0, 500, 200, 1000.0)
            idx.record("agent2", "s1", "gpt-4", "openai", 0.03, 80.0, 300, 100, 1001.0)
            assert abs(idx.total_cost("agent1", None, None) - 0.05) < 0.001
        except ImportError:
            pytest.skip("Native module not available")

    def test_group_by_provider(self):
        try:
            from astromesh._native import RustCostIndex
            idx = RustCostIndex()
            idx.record("a1", "s1", "gpt-4", "openai", 0.05, 100.0, 500, 200, 1000.0)
            idx.record("a1", "s1", "llama", "ollama", 0.01, 50.0, 100, 50, 1001.0)
            groups = idx.group_by(None, "provider")
            assert "openai" in groups
            assert "ollama" in groups
            assert groups["openai"][1] == 1  # calls count
        except ImportError:
            pytest.skip("Native module not available")



"""End-to-end integration test simulating what Forge SPA does."""

import pytest
import yaml
from asgi_lifespan import LifespanManager
from httpx import ASGITransport, AsyncClient

from astromesh.api.main import app


@pytest.fixture
def forge_config_dir(tmp_path, monkeypatch):
    config_dir = tmp_path / "config"
    config_dir.mkdir()
    (config_dir / "agents").mkdir()
    templates_dir = config_dir / "templates"
    templates_dir.mkdir()

    tpl = {
        "apiVersion": "astromesh/v1",
        "kind": "AgentTemplate",
        "metadata": {"name": "test-tpl", "version": "1.0.0", "category": "test", "tags": []},
        "template": {
            "display_name": "Test Template",
            "description": "A test template",
            "recommended_channels": [],
            "variables": [{"key": "name", "label": "Name", "required": True}],
            "agent_config": {
                "apiVersion": "astromesh/v1",
                "kind": "Agent",
                "metadata": {
                    "name": "test-agent",
                    "version": "1.0.0",
                    "namespace": "test",
                },
                "spec": {
                    "identity": {
                        "display_name": "Test Agent",
                        "description": "Test",
                    },
                    "model": {
                        "primary": {
                            "provider": "ollama",
                            "model": "test",
                            "endpoint": "http://localhost:11434",
                        },
                        "routing": {"strategy": "cost_optimized"},
                    },
                    "prompts": {"system": "You are a test agent."},
                    "orchestration": {"pattern": "react", "max_iterations": 5},
                },
            },
        },
    }
    (templates_dir / "test-tpl.template.yaml").write_text(yaml.dump(tpl))

    monkeypatch.setenv("ASTROMESH_CONFIG_DIR", str(config_dir))
    yield config_dir


async def test_full_forge_flow(forge_config_dir):
    """Simulates: list templates -> get template -> create agent -> deploy -> update -> delete."""
    async with LifespanManager(app):
        async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as client:
            # 1. List templates
            resp = await client.get("/v1/templates")
            assert resp.status_code == 200
            templates = resp.json()
            names = {t["name"] for t in templates}
            assert "test-tpl" in names

            # 2. Get template detail
            resp = await client.get("/v1/templates/test-tpl")
            assert resp.status_code == 200
            detail = resp.json()
            assert "agent_config" in detail

            # 3. Create agent (Forge resolves variables client-side)
            agent_config = detail["agent_config"]
            agent_config["metadata"]["name"] = "acme-agent"
            agent_config["spec"]["identity"]["display_name"] = "Acme Agent"
            resp = await client.post("/v1/agents", json=agent_config)
            assert resp.status_code == 201

            # 4. List agents — should show as draft
            resp = await client.get("/v1/agents")
            assert resp.status_code == 200
            agents = resp.json()["agents"]
            agent = next(a for a in agents if a["name"] == "acme-agent")
            assert agent["status"] == "draft"

            # 5. Update agent
            agent_config["spec"]["prompts"]["system"] = "Updated prompt"
            resp = await client.put("/v1/agents/acme-agent", json=agent_config)
            assert resp.status_code == 200

            # 6. Delete agent
            resp = await client.delete("/v1/agents/acme-agent")
            assert resp.status_code == 200

            # 7. Verify deleted
            resp = await client.get("/v1/agents")
            agents = resp.json()["agents"]
            assert not any(a["name"] == "acme-agent" for a in agents)



# tests/test_workflow_models.py
import pytest
from astromesh.workflow.models import (
    WorkflowSpec,
    StepSpec,
    StepResult,
    RetryConfig,
    WorkflowRunResult,
    StepType,
    StepStatus,
)


class TestRetryConfig:
    def test_defaults(self):
        rc = RetryConfig()
        assert rc.max_attempts == 1
        assert rc.backoff == "fixed"
        assert rc.initial_delay_seconds == 1.0

    def test_exponential(self):
        rc = RetryConfig(max_attempts=3, backoff="exponential", initial_delay_seconds=2.0)
        assert rc.max_attempts == 3
        assert rc.backoff == "exponential"


class TestStepSpec:
    def test_agent_step(self):
        step = StepSpec(
            name="research",
            agent="web-researcher",
            input_template="{{ trigger.query }}",
        )
        assert step.step_type == StepType.AGENT
        assert step.agent == "web-researcher"

    def test_tool_step(self):
        step = StepSpec(
            name="store",
            tool="cache_store",
            arguments={"key": "k1", "value": "v1"},
        )
        assert step.step_type == StepType.TOOL

    def test_switch_step(self):
        step = StepSpec(
            name="decide",
            switch=[
                {"when": "{{ steps.qualify.output.data.score > 7 }}", "goto": "send-email"},
                {"default": True, "goto": "log-skip"},
            ],
        )
        assert step.step_type == StepType.SWITCH

    def test_invalid_step_raises(self):
        """A step with no agent, tool, or switch is invalid."""
        with pytest.raises(ValueError, match="must have exactly one"):
            StepSpec(name="broken")

    def test_retry_config_from_dict(self):
        step = StepSpec(
            name="s1",
            agent="a1",
            input_template="hi",
            retry={"max_attempts": 3, "backoff": "exponential", "initial_delay_seconds": 2.0},
        )
        assert step.retry.max_attempts == 3
        assert step.retry.backoff == "exponential"


class TestStepResult:
    def test_success(self):
        sr = StepResult(name="research", status=StepStatus.SUCCESS, output={"answer": "hello"})
        assert sr.status == StepStatus.SUCCESS
        assert sr.output["answer"] == "hello"
        assert sr.error is None

    def test_error(self):
        sr = StepResult(name="research", status=StepStatus.ERROR, error="timeout")
        assert sr.status == StepStatus.ERROR


class TestWorkflowSpec:
    def test_basic_workflow(self):
        wf = WorkflowSpec(
            name="test-wf",
            trigger="api",
            timeout_seconds=300,
            steps=[
                StepSpec(name="s1", agent="a1", input_template="{{ trigger.query }}"),
            ],
        )
        assert wf.name == "test-wf"
        assert wf.trigger == "api"
        assert len(wf.steps) == 1

    def test_step_lookup_by_name(self):
        wf = WorkflowSpec(
            name="test-wf",
            trigger="api",
            steps=[
                StepSpec(name="s1", agent="a1", input_template="hi"),
                StepSpec(name="s2", agent="a2", input_template="bye"),
            ],
        )
        assert wf.get_step("s1").agent == "a1"
        assert wf.get_step("s2").agent == "a2"
        assert wf.get_step("nonexistent") is None

    def test_duplicate_step_names_raises(self):
        with pytest.raises(ValueError, match="[Dd]uplicate"):
            WorkflowSpec(
                name="test-wf",
                trigger="api",
                steps=[
                    StepSpec(name="s1", agent="a1", input_template="hi"),
                    StepSpec(name="s1", agent="a2", input_template="bye"),
                ],
            )


class TestWorkflowRunResult:
    def test_basic(self):
        result = WorkflowRunResult(
            workflow_name="test-wf",
            status="completed",
            steps={"s1": StepResult(name="s1", status=StepStatus.SUCCESS, output={"answer": "ok"})},
            output={"answer": "ok"},
        )
        assert result.workflow_name == "test-wf"
        assert result.status == "completed"
        assert "s1" in result.steps



import json
import uuid

import asyncpg

from astromesh.core.memory import SemanticBackend, SemanticMemory


class PGVectorSemanticBackend(SemanticBackend):
    def __init__(
        self, pool: asyncpg.Pool | None = None, dsn: str | None = None, dimension: int = 1536
    ):
        self._pool = pool
        self._dsn = dsn
        self._dimension = dimension

    async def initialize(self):
        if self._pool is None and self._dsn:
            self._pool = await asyncpg.create_pool(self._dsn)
        async with self._pool.acquire() as conn:
            await conn.execute("CREATE EXTENSION IF NOT EXISTS vector")
            await conn.execute(
                f"CREATE TABLE IF NOT EXISTS semantic_memories ("
                f"id TEXT PRIMARY KEY, "
                f"agent_id TEXT, "
                f"content TEXT, "
                f"embedding vector({self._dimension}), "
                f"metadata JSONB DEFAULT '{{}}'"
                f")"
            )

    async def store(self, agent_id, content, embedding, metadata):
        memory_id = str(uuid.uuid4())
        async with self._pool.acquire() as conn:
            await conn.execute(
                "INSERT INTO semantic_memories (id, agent_id, content, embedding, metadata) "
                "VALUES ($1, $2, $3, $4, $5)",
                memory_id,
                agent_id,
                content,
                str(embedding),
                json.dumps(metadata),
            )
        return memory_id

    async def search(self, agent_id, query_embedding, top_k=10, threshold=0.7):
        async with self._pool.acquire() as conn:
            rows = await conn.fetch(
                "SELECT id, content, metadata, "
                "1 - (embedding <=> $1::vector) AS similarity "
                "FROM semantic_memories "
                "WHERE agent_id = $2 AND 1 - (embedding <=> $1::vector) >= $3 "
                "ORDER BY similarity DESC LIMIT $4",
                str(query_embedding),
                agent_id,
                threshold,
                top_k,
            )
        return [
            SemanticMemory(
                content=row["content"],
                embedding=[],
                metadata=json.loads(row["metadata"])
                if isinstance(row["metadata"], str)
                else row["metadata"],
                similarity=float(row["similarity"]),
                source=row["id"],
            )
            for row in rows
        ]

    async def delete(self, agent_id, memory_id):
        async with self._pool.acquire() as conn:
            await conn.execute(
                "DELETE FROM semantic_memories WHERE id = $1 AND agent_id = $2",
                memory_id,
                agent_id,
            )



import json
from datetime import datetime, timezone
from zoneinfo import ZoneInfo

from jinja2 import BaseLoader, Environment, TemplateSyntaxError, UndefinedError

from astromesh.tools.base import BuiltinTool, ToolContext, ToolResult


class DatetimeNowTool(BuiltinTool):
    name = "datetime_now"
    description = "Get the current date and time with optional timezone"
    parameters = {
        "type": "object",
        "properties": {
            "timezone": {
                "type": "string",
                "description": "Timezone name (e.g. 'UTC', 'US/Eastern'). Defaults to UTC.",
            }
        },
    }

    async def execute(self, arguments: dict, context: ToolContext) -> ToolResult:
        tz_name = arguments.get("timezone", "UTC")
        try:
            tz = ZoneInfo(tz_name)
        except (KeyError, Exception):
            tz = timezone.utc
            tz_name = "UTC"
        now = datetime.now(tz)
        return ToolResult(
            success=True,
            data={
                "datetime": now.isoformat(),
                "timezone": tz_name,
                "unix_timestamp": now.timestamp(),
            },
            metadata={},
        )


class JsonTransformTool(BuiltinTool):
    name = "json_transform"
    description = "Transform JSON data using a Jinja2 template that outputs JSON"
    parameters = {
        "type": "object",
        "properties": {
            "data": {"description": "The input data to transform"},
            "template": {
                "type": "string",
                "description": "Jinja2 template that produces valid JSON",
            },
        },
        "required": ["data", "template"],
    }

    async def execute(self, arguments: dict, context: ToolContext) -> ToolResult:
        data = arguments["data"]
        template_str = arguments["template"]
        try:
            env = Environment(loader=BaseLoader())
            template = env.from_string(template_str)
            rendered = template.render(data=data)
            parsed = json.loads(rendered)
            return ToolResult(success=True, data=parsed, metadata={})
        except (TemplateSyntaxError, UndefinedError, json.JSONDecodeError) as e:
            return ToolResult(success=False, data=None, metadata={}, error=str(e))


class CacheStoreTool(BuiltinTool):
    name = "cache_store"
    description = "Temporary key-value cache for sharing data between tool calls"
    parameters = {
        "type": "object",
        "properties": {
            "action": {
                "type": "string",
                "enum": ["get", "set", "delete"],
                "description": "Cache operation",
            },
            "key": {"type": "string", "description": "Cache key"},
            "value": {"description": "Value to store (for set action)"},
        },
        "required": ["action", "key"],
    }

    async def execute(self, arguments: dict, context: ToolContext) -> ToolResult:
        action = arguments["action"]
        key = arguments["key"]
        cache = context.cache
        if action == "set":
            cache[key] = arguments.get("value")
            return ToolResult(success=True, data=None, metadata={"action": "set", "key": key})
        elif action == "get":
            return ToolResult(
                success=True,
                data=cache.get(key),
                metadata={"action": "get", "key": key},
            )
        elif action == "delete":
            cache.pop(key, None)
            return ToolResult(success=True, data=None, metadata={"action": "delete", "key": key})
        return ToolResult(success=False, data=None, error=f"Unknown action: {action}")



import pytest
from astromesh.runtime.engine import AgentRuntime


def _make_config(name: str, version: str = "0.1.0", namespace: str = "test") -> dict:
    """Build a minimal agent config dict matching the schema expected by _build_agent()."""
    return {
        "apiVersion": "astromesh/v1",
        "kind": "Agent",
        "metadata": {
            "name": name,
            "version": version,
            "namespace": namespace,
        },
        "spec": {
            "identity": {
                "display_name": f"{name} display",
                "description": f"Dynamic test agent: {name}",
            },
            "model": {
                "primary": {
                    "provider": "ollama",
                    "model": "llama3:8b",
                    "endpoint": "http://ollama:11434",
                },
                "routing": {"strategy": "cost_optimized"},
            },
            "prompts": {
                "system": "You are a dynamic test agent.",
            },
            "orchestration": {
                "pattern": "react",
                "max_iterations": 5,
            },
        },
    }


@pytest.fixture
def runtime(tmp_path):
    """An AgentRuntime with an empty (but valid) config dir — no bootstrap needed."""
    agents_dir = tmp_path / "agents"
    agents_dir.mkdir()
    return AgentRuntime(config_dir=str(tmp_path))


async def test_register_agent_adds_to_configs(runtime):
    config = _make_config("dynamic-agent")
    await runtime.register_agent(config)

    # register_agent stores in _agent_configs with status 'draft' (not in _agents)
    assert "dynamic-agent" in runtime._agent_configs
    assert runtime._agent_status["dynamic-agent"] == "draft"


async def test_register_agent_upsert_overwrites(runtime):
    config_v1 = _make_config("my-agent", version="0.1.0")
    config_v2 = _make_config("my-agent", version="0.2.0")

    await runtime.register_agent(config_v1)
    assert runtime._agent_configs["my-agent"]["metadata"]["version"] == "0.1.0"

    # Second register must not raise — upsert semantics
    await runtime.register_agent(config_v2)
    assert runtime._agent_configs["my-agent"]["metadata"]["version"] == "0.2.0"


async def test_unregister_agent_removes(runtime):
    config = _make_config("to-remove")
    await runtime.register_agent(config)
    assert "to-remove" in runtime._agent_configs

    runtime.unregister_agent("to-remove")
    assert "to-remove" not in runtime._agent_configs
    assert "to-remove" not in runtime._agent_status


async def test_unregister_agent_not_found_raises(runtime):
    with pytest.raises(ValueError, match="not found"):
        runtime.unregister_agent("ghost-agent")



"""Astromesh ADK CLI."""

from __future__ import annotations

import asyncio
import importlib
import importlib.util
import sys
from pathlib import Path

import typer
from rich.console import Console
from rich.table import Table

app = typer.Typer(name="astromesh-adk", help="Astromesh Agent Development Kit CLI")
console = Console()


def _load_module(file_path: str):
    """Import a Python file as a module."""
    path = Path(file_path).resolve()
    if not path.exists():
        console.print(f"[red]File not found: {path}[/red]")
        raise typer.Exit(1)

    spec = importlib.util.spec_from_file_location("__adk_user_module__", path)
    module = importlib.util.module_from_spec(spec)
    sys.modules["__adk_user_module__"] = module
    spec.loader.exec_module(module)
    return module


def _discover_agents(module):
    """Find all AgentWrapper and Agent instances in a module."""
    from astromesh_adk.agent import AgentWrapper, Agent

    agents = []
    for name in dir(module):
        obj = getattr(module, name)
        if isinstance(obj, AgentWrapper):
            agents.append(obj)
        elif isinstance(obj, type) and issubclass(obj, Agent) and obj is not Agent:
            agents.append(obj())
    return agents


def _parse_agent_ref(ref: str):
    """Parse 'file.py:agent_name' into (file_path, agent_name)."""
    if ":" in ref:
        file_path, agent_name = ref.rsplit(":", 1)
        return file_path, agent_name
    return ref, None


@app.command()
def list(file: str = typer.Argument(..., help="Python file with agent definitions")):
    """List all agents defined in a file."""
    module = _load_module(file)
    agents = _discover_agents(module)

    if not agents:
        console.print("[yellow]No agents found.[/yellow]")
        raise typer.Exit(0)

    table = Table(title="Agents")
    table.add_column("Name", style="cyan")
    table.add_column("Model", style="green")
    table.add_column("Pattern", style="blue")
    table.add_column("Tools", style="yellow")

    for a in agents:
        tool_count = str(len(a.tools))
        table.add_row(a.name, a.model, a.pattern, tool_count)

    console.print(table)


@app.command()
def run(
    agent_ref: str = typer.Argument(..., help="agent file:name (e.g., agents.py:my_agent)"),
    query: str = typer.Argument(..., help="Query to send to the agent"),
    session: str = typer.Option("default", "--session", "-s", help="Session ID"),
):
    """Run an agent with a query."""
    file_path, agent_name = _parse_agent_ref(agent_ref)
    module = _load_module(file_path)
    agents = _discover_agents(module)

    target = None
    for a in agents:
        if agent_name is None or a.name == agent_name:
            target = a
            break

    if target is None:
        console.print(f"[red]Agent '{agent_name}' not found.[/red]")
        raise typer.Exit(1)

    async def _run():
        result = await target.run(query, session_id=session)
        console.print(f"\n[bold green]Answer:[/bold green] {result.answer}")
        if result.steps:
            console.print(f"\n[dim]Steps: {len(result.steps)} | Cost: ${result.cost:.4f} | Tokens: {result.tokens}[/dim]")

    asyncio.run(_run())


@app.command()
def chat(
    agent_ref: str = typer.Argument(..., help="agent file:name"),
    session: str = typer.Option("default", "--session", "-s", help="Session ID"),
):
    """Interactive chat with an agent."""
    file_path, agent_name = _parse_agent_ref(agent_ref)
    module = _load_module(file_path)
    agents = _discover_agents(module)

    target = None
    for a in agents:
        if agent_name is None or a.name == agent_name:
            target = a
            break

    if target is None:
        console.print(f"[red]Agent '{agent_name}' not found.[/red]")
        raise typer.Exit(1)

    console.print(f"[bold]Chat with {target.name}[/bold] (Ctrl+C to exit)\n")

    async def _chat():
        while True:
            try:
                user_input = console.input("[bold blue]> [/bold blue]")
                if not user_input.strip():
                    continue
                result = await target.run(user_input, session_id=session)
                console.print(f"\n{result.answer}\n")
            except KeyboardInterrupt:
                console.print("\n[dim]Bye![/dim]")
                break

    asyncio.run(_chat())


@app.command()
def check(file: str = typer.Argument(..., help="Python file to validate")):
    """Validate agent definitions."""
    import os
    module = _load_module(file)
    agents = _discover_agents(module)

    if not agents:
        console.print("[yellow]No agents found.[/yellow]")
        raise typer.Exit(0)

    from astromesh_adk.providers import parse_model_string, PROVIDER_REGISTRY

    for a in agents:
        issues = []

        # Check provider env vars
        provider_name, _ = parse_model_string(a.model)
        entry = PROVIDER_REGISTRY.get(provider_name)
        if entry and entry.get("env_var"):
            if not os.environ.get(entry["env_var"]):
                issues.append(f"Missing env var: {entry['env_var']}")

        # Check tools
        for t in a.tools:
            from astromesh_adk.tools import ToolDefinitionWrapper, Tool
            from astromesh_adk.mcp import MCPToolSet
            if isinstance(t, ToolDefinitionWrapper):
                pass  # OK
            elif isinstance(t, Tool):
                if not t.name:
                    issues.append("Tool class missing 'name'")
            elif isinstance(t, MCPToolSet):
                import shutil
                cmd = t.config.get("command")
                if cmd and not shutil.which(cmd):
                    issues.append(f"MCP command not found: {cmd}")

        status = "[green]v[/green]" if not issues else "[red]x[/red]"
        tool_count = len(a.tools)
        console.print(f"  {status} {a.name}: {provider_name} provider, {tool_count} tools")
        for issue in issues:
            console.print(f"    [yellow]! {issue}[/yellow]")


@app.command()
def dev(
    file: str = typer.Argument(..., help="Python file with agent definitions"),
    port: int = typer.Option(8000, "--port", "-p", help="Server port"),
    reload: bool = typer.Option(False, "--reload", "-r", help="Enable hot reload"),
):
    """Start the development server."""
    console.print("[bold]Starting ADK dev server...[/bold]")
    console.print(f"  File: {file}")
    console.print(f"  Port: {port}")
    console.print(f"  Reload: {reload}")

    # Import here to avoid requiring FastAPI as a core dependency
    try:
        import uvicorn
        from fastapi import FastAPI
    except ImportError:
        console.print("[red]Dev server requires FastAPI and Uvicorn. Install with: pip install astromesh-adk[dev][/red]")
        raise typer.Exit(1)

    module = _load_module(file)
    agents = _discover_agents(module)

    if not agents:
        console.print("[yellow]No agents found.[/yellow]")
        raise typer.Exit(0)

    # Build a minimal FastAPI app
    dev_app = FastAPI(title="ADK Dev Server")

    @dev_app.get("/v1/health")
    def health():
        return {"status": "ok", "agents": [a.name for a in agents]}

    @dev_app.get("/v1/agents")
    def list_agents():
        return [{"name": a.name, "model": a.model, "description": a.description} for a in agents]

    @dev_app.post("/v1/agents/{agent_name}/run")
    async def run_agent(agent_name: str, body: dict):
        target = next((a for a in agents if a.name == agent_name), None)
        if not target:
            return {"error": f"Agent '{agent_name}' not found"}
        result = await target.run(
            body.get("query", ""),
            session_id=body.get("session_id", "default"),
            context=body.get("context"),
        )
        return {"answer": result.answer, "steps": result.steps}

    console.print(f"\n[bold green]Serving {len(agents)} agent(s) at http://localhost:{port}[/bold green]")
    for a in agents:
        console.print(f"  - {a.name} ({a.model})")

    uvicorn.run(dev_app, host="0.0.0.0", port=port, reload=reload)
